"""fileWrite.py unit tests."""
import io
from unittest.mock import mock_open, patch

import pytest

from pypyr.context import Context
from pypyr.errors import (
    ContextError,
    KeyInContextHasNoValueError,
    KeyNotInContextError)
import pypyr.steps.filewrite as filewrite

# region validation


def test_filewrite_no_input_raises():
    """No input fileWrite raises."""
    context = Context({
        'k1': 'v1'})

    with pytest.raises(KeyNotInContextError) as err_info:
        filewrite.run_step(context)

    assert str(err_info.value) == (
        "context['fileWrite'] doesn't exist. "
        "It must exist for pypyr.steps.filewrite.")


def test_filewrite_none_filewrite_raises():
    """None fileWrite raises."""
    context = Context({
        'k1': 'v1',
        'fileWrite': None})

    with pytest.raises(KeyInContextHasNoValueError) as err_info:
        filewrite.run_step(context)

    assert str(err_info.value) == (
        "context['fileWrite'] must have a value for "
        "pypyr.steps.filewrite.")


def test_filewrite_filewrite_not_iterable_raises():
    """Not iterable fileWrite raises."""
    context = Context({
        'k1': 'v1',
        'fileWrite': 1})

    with pytest.raises(ContextError) as err_info:
        filewrite.run_step(context)

    assert str(err_info.value) == (
        "context['fileWrite'] must exist, be iterable "
        "and contain 'path' for pypyr.steps.filewrite. argument of type "
        "'int' is not iterable")


def test_filewrite_empty_path_raises():
    """Empty path raises."""
    context = Context({
        'fileWrite': {
            'path': ''
        }})

    with pytest.raises(KeyInContextHasNoValueError) as err_info:
        filewrite.run_step(context)

    assert str(err_info.value) == (
        "context['fileWrite']['path'] must have a value for "
        "pypyr.steps.filewrite.")


def test_filewrite_none_path_raises():
    """Empty path raises."""
    context = Context({
        'fileWrite': {
            'path': None
        }})

    with pytest.raises(KeyInContextHasNoValueError) as err_info:
        filewrite.run_step(context)

    assert str(err_info.value) == (
        "context['fileWrite']['path'] must have a value for "
        "pypyr.steps.filewrite.")


def test_filewrite_no_path_raises():
    """No path raises."""
    context = Context({
        'fileWrite': 'blah',
        'k1': 'v1'})

    with pytest.raises(KeyNotInContextError) as err_info:
        filewrite.run_step(context)

    assert str(err_info.value) == ("context['fileWrite']['path'] "
                                   "doesn't exist. It must exist for "
                                   "pypyr.steps.filewrite.")


def test_filewrite_no_payload_raises():
    """No payload raises."""
    context = Context({
        'fileWrite': {
            'path': 'arb'
        }})

    with pytest.raises(KeyNotInContextError) as err_info:
        filewrite.run_step(context)

    assert str(err_info.value) == ("context['fileWrite']['payload'] "
                                   "doesn't exist. It must exist for "
                                   "pypyr.steps.filewrite.")
# endregion validation

# region write text


@patch('pypyr.steps.filewrite.Path')
def test_filewrite_pass_with_payload(mock_path):
    """Success case writes only specific context payload."""
    context = Context({
        'k1': 'v1',
        'fileWrite': {
            'path': '/arb/blah',
            'payload': "one\ntwo\nthree"
        }})

    with io.StringIO() as out_text:
        with patch('pypyr.steps.filewrite.open',
                   mock_open()) as mock_output:
            mock_output.return_value.write.side_effect = out_text.write
            filewrite.run_step(context)

        payload = out_text.getvalue()

    assert context, "context shouldn't be None"
    assert len(context) == 2, "context should have 2 items"
    assert context['k1'] == 'v1'
    assert context['fileWrite'] == {'path': '/arb/blah',
                                    'payload': "one\ntwo\nthree"}

    mock_path.assert_called_once_with('/arb/blah')
    mocked_path = mock_path.return_value
    mocked_path.parent.mkdir.assert_called_once_with(parents=True,
                                                     exist_ok=True)
    mock_output.assert_called_once_with(mocked_path, 'w', encoding=None)

    assert payload == 'one\ntwo\nthree'


@patch('pypyr.steps.filewrite.Path')
def test_filewrite_pass_with_payload_encoding(mock_path):
    """Success case writes only specific context payload and encoding."""
    context = Context({
        'k1': 'v1',
        'fileWrite': {
            'path': '/arb/blah',
            'payload': 'one\ntwo\nthree',
            'encoding': 'arb'
        }})

    with io.StringIO() as out_text:
        with patch('pypyr.steps.filewrite.open',
                   mock_open()) as mock_output:
            mock_output.return_value.write.side_effect = out_text.write
            filewrite.run_step(context)

        payload = out_text.getvalue()

    assert context, "context shouldn't be None"
    assert len(context) == 2, "context should have 2 items"
    assert context['k1'] == 'v1'
    assert context['fileWrite'] == {'path': '/arb/blah',
                                    'payload': 'one\ntwo\nthree',
                                    'encoding': 'arb'}

    mock_path.assert_called_once_with('/arb/blah')
    mocked_path = mock_path.return_value
    mocked_path.parent.mkdir.assert_called_once_with(parents=True,
                                                     exist_ok=True)
    mock_output.assert_called_once_with(mocked_path, 'w', encoding='arb')

    assert payload == 'one\ntwo\nthree'


@patch('pypyr.steps.filewrite.Path')
def test_filewrite_pass_encoding_from_config_override(mock_path,
                                                      monkeypatch):
    """Direct encoding overrides config encoding.."""
    monkeypatch.setattr('pypyr.config.config.default_encoding', 'arbx')
    context = Context({
        'k1': 'v1',
        'fileWrite': {
            'path': '/arb/blah',
            'payload': 'one\ntwo\nthree',
            'encoding': 'arb'
        }})

    with io.StringIO() as out_text:
        with patch('pypyr.steps.filewrite.open',
                   mock_open()) as mock_output:
            mock_output.return_value.write.side_effect = out_text.write
            filewrite.run_step(context)

        payload = out_text.getvalue()

    assert context, "context shouldn't be None"
    assert len(context) == 2, "context should have 2 items"
    assert context['k1'] == 'v1'
    assert context['fileWrite'] == {'path': '/arb/blah',
                                    'payload': 'one\ntwo\nthree',
                                    'encoding': 'arb'}

    mock_path.assert_called_once_with('/arb/blah')
    mocked_path = mock_path.return_value
    mocked_path.parent.mkdir.assert_called_once_with(parents=True,
                                                     exist_ok=True)
    mock_output.assert_called_once_with(mocked_path, 'w', encoding='arb')

    assert payload == 'one\ntwo\nthree'


@patch('pypyr.steps.filewrite.Path')
def test_filewrite_pass_encoding_from_config(mock_path,
                                             monkeypatch):
    """Encoding comes from config."""
    monkeypatch.setattr('pypyr.config.config.default_encoding', 'arbx')
    context = Context({
        'k1': 'v1',
        'fileWrite': {
            'path': '/arb/blah',
            'payload': 'one\ntwo\nthree'
        }})

    with io.StringIO() as out_text:
        with patch('pypyr.steps.filewrite.open',
                   mock_open()) as mock_output:
            mock_output.return_value.write.side_effect = out_text.write
            filewrite.run_step(context)

        payload = out_text.getvalue()

    assert context, "context shouldn't be None"
    assert len(context) == 2, "context should have 2 items"
    assert context['k1'] == 'v1'
    assert context['fileWrite'] == {'path': '/arb/blah',
                                    'payload': 'one\ntwo\nthree'}

    mock_path.assert_called_once_with('/arb/blah')
    mocked_path = mock_path.return_value
    mocked_path.parent.mkdir.assert_called_once_with(parents=True,
                                                     exist_ok=True)
    mock_output.assert_called_once_with(mocked_path, 'w', encoding='arbx')

    assert payload == 'one\ntwo\nthree'


@patch('pypyr.steps.filewrite.Path')
def test_filewrite_pass_with_non_string_payload(mock_path):
    """Empty payload write non string types to file."""
    context = Context({
        'k1': 'v1',
        'fileWrite': {
            'path': '/arb/blah',
            'payload': 123
        }})

    with io.StringIO() as out_text:
        with patch('pypyr.steps.filewrite.open',
                   mock_open()) as mock_output:
            mock_output.return_value.write.side_effect = out_text.write
            filewrite.run_step(context)

        payload = out_text.getvalue()

    mock_path.assert_called_once_with('/arb/blah')
    mocked_path = mock_path.return_value
    mocked_path.parent.mkdir.assert_called_once_with(parents=True,
                                                     exist_ok=True)
    mock_output.assert_called_once_with(mocked_path, 'w', encoding=None)

    assert payload == '123'


@patch('pypyr.steps.filewrite.Path')
def test_filewrite_pass_with_empty_payload(mock_path):
    """Empty payload write empty file."""
    context = Context({
        'k1': 'v1',
        'fileWrite': {
            'path': '/arb/blah',
            'payload': ''
        }})

    with io.StringIO() as out_text:
        with patch('pypyr.steps.filewrite.open',
                   mock_open()) as mock_output:
            mock_output.return_value.write.side_effect = out_text.write
            filewrite.run_step(context)

        payload = out_text.getvalue()

    mock_path.assert_called_once_with('/arb/blah')
    mocked_path = mock_path.return_value
    mocked_path.parent.mkdir.assert_called_once_with(parents=True,
                                                     exist_ok=True)
    mock_output.assert_called_once_with(mocked_path, 'w', encoding=None)

    assert payload == ''


@patch('pypyr.steps.filewrite.Path')
def test_filewrite_pass_with_none_payload(mock_path):
    """None payload write empty file."""
    context = Context({
        'k1': 'v1',
        'fileWrite': {
            'path': '/arb/blah',
            'payload': None
        }})

    with io.StringIO() as out_text:
        with patch('pypyr.steps.filewrite.open',
                   mock_open()) as mock_output:
            mock_output.return_value.write.side_effect = out_text.write
            filewrite.run_step(context)

        payload = out_text.getvalue()

    mock_path.assert_called_once_with('/arb/blah')
    mocked_path = mock_path.return_value
    mocked_path.parent.mkdir.assert_called_once_with(parents=True,
                                                     exist_ok=True)
    mock_output.assert_called_once_with(mocked_path, 'w', encoding=None)

    assert payload == 'None'

# endregion write text
# region append


@patch('pypyr.steps.filewrite.Path')
def test_filewrite_append(mock_path):
    """Append to file."""
    context = Context({
        'k1': 'v1',
        'fileWrite': {
            'path': '/arb/blah',
            'payload': "one\ntwo\nthree",
            'append': True,
        }})

    with io.StringIO() as out_text:
        with patch('pypyr.steps.filewrite.open',
                   mock_open()) as mock_output:
            mock_output.return_value.write.side_effect = out_text.write
            filewrite.run_step(context)

        payload = out_text.getvalue()

    mock_path.assert_called_once_with('/arb/blah')
    mocked_path = mock_path.return_value
    mocked_path.parent.mkdir.assert_called_once_with(parents=True,
                                                     exist_ok=True)
    mock_output.assert_called_once_with(mocked_path, 'a', encoding=None)

    assert payload == 'one\ntwo\nthree'


@patch('pypyr.steps.filewrite.Path')
def test_filewrite_append_false(mock_path):
    """Open in write mode when append explicitly False."""
    context = Context({
        'k1': 'v1',
        'fileWrite': {
            'path': '/arb/blah',
            'payload': "one\ntwo\nthree",
            'append': False,
        }})

    with io.StringIO() as out_text:
        with patch('pypyr.steps.filewrite.open',
                   mock_open()) as mock_output:
            mock_output.return_value.write.side_effect = out_text.write
            filewrite.run_step(context)

        payload = out_text.getvalue()

    mock_path.assert_called_once_with('/arb/blah')
    mocked_path = mock_path.return_value
    mocked_path.parent.mkdir.assert_called_once_with(parents=True,
                                                     exist_ok=True)
    mock_output.assert_called_once_with(mocked_path, 'w', encoding=None)

    assert payload == 'one\ntwo\nthree'

# endregion append

# region binary


@patch('pypyr.steps.filewrite.Path')
def test_filewrite_binary(mock_path):
    """Write binary data."""
    context = Context({
        'k1': 'v1',
        'fileWrite': {
            'path': '/arb/blah',
            'payload': b"one\ntwo\nthree",
            'binary': True,
        }})

    with io.BytesIO() as out_bytes:
        with patch('pypyr.steps.filewrite.open',
                   mock_open()) as mock_output:
            mock_output.return_value.write.side_effect = out_bytes.write
            filewrite.run_step(context)

        payload = out_bytes.getvalue()

    mock_path.assert_called_once_with('/arb/blah')
    mocked_path = mock_path.return_value
    mocked_path.parent.mkdir.assert_called_once_with(parents=True,
                                                     exist_ok=True)
    mock_output.assert_called_once_with(mocked_path, 'wb', encoding=None)

    assert payload == b'one\ntwo\nthree'


@patch('pypyr.steps.filewrite.Path')
def test_filewrite_binary_false(mock_path):
    """Open in write mode when binary explicitly False."""
    context = Context({
        'k1': 'v1',
        'fileWrite': {
            'path': '/arb/blah',
            'payload': "one\ntwo\nthree",
            'binary': False,
        }})

    with io.StringIO() as out_text:
        with patch('pypyr.steps.filewrite.open',
                   mock_open()) as mock_output:
            mock_output.return_value.write.side_effect = out_text.write
            filewrite.run_step(context)

        payload = out_text.getvalue()

    mock_path.assert_called_once_with('/arb/blah')
    mocked_path = mock_path.return_value
    mocked_path.parent.mkdir.assert_called_once_with(parents=True,
                                                     exist_ok=True)
    mock_output.assert_called_once_with(mocked_path, 'w', encoding=None)

    assert payload == 'one\ntwo\nthree'

# endregion binary

# region binary + append


@patch('pypyr.steps.filewrite.Path')
def test_filewrite_binary_append(mock_path):
    """Append binary data."""
    context = Context({
        'k1': 'v1',
        'fileWrite': {
            'path': '/arb/blah',
            'payload': b"one\ntwo\nthree",
            'binary': True,
            'append': True
        }})

    with io.BytesIO() as out_bytes:
        with patch('pypyr.steps.filewrite.open',
                   mock_open()) as mock_output:
            mock_output.return_value.write.side_effect = out_bytes.write
            filewrite.run_step(context)

        payload = out_bytes.getvalue()

    mock_path.assert_called_once_with('/arb/blah')
    mocked_path = mock_path.return_value
    mocked_path.parent.mkdir.assert_called_once_with(parents=True,
                                                     exist_ok=True)
    mock_output.assert_called_once_with(mocked_path, 'ab', encoding=None)

    assert payload == b'one\ntwo\nthree'


@patch('pypyr.steps.filewrite.Path')
def test_filewrite_binary_true_append_false(mock_path):
    """Write binary data with append explicit False."""
    context = Context({
        'k1': 'v1',
        'fileWrite': {
            'path': '/arb/blah',
            'payload': b"one\ntwo\nthree",
            'binary': True,
            'append': False
        }})

    with io.BytesIO() as out_bytes:
        with patch('pypyr.steps.filewrite.open',
                   mock_open()) as mock_output:
            mock_output.return_value.write.side_effect = out_bytes.write
            filewrite.run_step(context)

        payload = out_bytes.getvalue()

    mock_path.assert_called_once_with('/arb/blah')
    mocked_path = mock_path.return_value
    mocked_path.parent.mkdir.assert_called_once_with(parents=True,
                                                     exist_ok=True)
    mock_output.assert_called_once_with(mocked_path, 'wb', encoding=None)

    assert payload == b'one\ntwo\nthree'


@patch('pypyr.steps.filewrite.Path')
def test_filewrite_binary_false_append_true(mock_path):
    """Open in append mode when binary explicitly False."""
    context = Context({
        'k1': 'v1',
        'fileWrite': {
            'path': '/arb/blah',
            'payload': "one\ntwo\nthree",
            'binary': False,
            'append': True
        }})

    with io.StringIO() as out_text:
        with patch('pypyr.steps.filewrite.open',
                   mock_open()) as mock_output:
            mock_output.return_value.write.side_effect = out_text.write
            filewrite.run_step(context)

        payload = out_text.getvalue()

    mock_path.assert_called_once_with('/arb/blah')
    mocked_path = mock_path.return_value
    mocked_path.parent.mkdir.assert_called_once_with(parents=True,
                                                     exist_ok=True)
    mock_output.assert_called_once_with(mocked_path, 'a', encoding=None)

    assert payload == 'one\ntwo\nthree'


@patch('pypyr.steps.filewrite.Path')
def test_filewrite_binary_false_append_false(mock_path):
    """Open in write mode when binary & append explicitly False."""
    context = Context({
        'k1': 'v1',
        'fileWrite': {
            'path': '/arb/blah',
            'payload': "one\ntwo\nthree",
            'binary': False,
            'append': False
        }})

    with io.StringIO() as out_text:
        with patch('pypyr.steps.filewrite.open',
                   mock_open()) as mock_output:
            mock_output.return_value.write.side_effect = out_text.write
            filewrite.run_step(context)

        payload = out_text.getvalue()

    mock_path.assert_called_once_with('/arb/blah')
    mocked_path = mock_path.return_value
    mocked_path.parent.mkdir.assert_called_once_with(parents=True,
                                                     exist_ok=True)
    mock_output.assert_called_once_with(mocked_path, 'w', encoding=None)

    assert payload == 'one\ntwo\nthree'

# endregion binary + append

# region substitutions


@patch('pypyr.steps.filewrite.Path')
def test_filewrite_pass_with_substitutions(mock_path):
    """Success case writes only specified context with substitutions."""
    context = Context({
        'k1': 'v1',
        'p': '/arb/path',
        'intkey': 3,
        'is_bin': True,
        'is_append': True,
        'enc': 'arb',
        'fileWrite': {
            'path': '{p}',
            'payload': 'begin {k1} {intkey} end',
            'binary': '{is_bin}',
            'append': '{is_append}',
            'encoding': '{enc}'
        }})

    with io.StringIO() as out_text:
        with patch('pypyr.steps.filewrite.open',
                   mock_open()) as mock_output:
            mock_output.return_value.write.side_effect = out_text.write
            filewrite.run_step(context)

        payload = out_text.getvalue()

    assert context, "context shouldn't be None"
    assert len(context) == 7, "context should have 7 items"
    assert context['k1'] == 'v1'
    assert context['fileWrite'] == {
        'path': '{p}',
        'payload': 'begin {k1} {intkey} end',
        'binary': '{is_bin}',
        'append': '{is_append}',
        'encoding': '{enc}'
    }

    mock_path.assert_called_once_with('/arb/path')
    mocked_path = mock_path.return_value
    mocked_path.parent.mkdir.assert_called_once_with(parents=True,
                                                     exist_ok=True)
    mock_output.assert_called_once_with(mocked_path, 'ab', encoding='arb')

    assert payload == 'begin v1 3 end'


@patch('pypyr.steps.filewrite.Path')
def test_filewrite_pass_with_substitutions_bare_payload(mock_path):
    """Write bare formatting expression as payload."""
    context = Context({
        'k1': 'v1',
        'p': '/arb/path',
        'out': 'one\ntwo\nthree',
        'is_append': False,
        'fileWrite': {
            'path': '{p}',
            'payload': '{out}',
            'append': '{is_append}'
        }})

    with io.StringIO() as out_text:
        with patch('pypyr.steps.filewrite.open',
                   mock_open()) as mock_output:
            mock_output.return_value.write.side_effect = out_text.write
            filewrite.run_step(context)

        payload = out_text.getvalue()

    assert context, "context shouldn't be None"
    assert len(context) == 5, "context should have 5 items"
    assert context['k1'] == 'v1'
    assert context['fileWrite'] == {
        'path': '{p}',
        'payload': '{out}',
        'append': '{is_append}'
    }

    mock_path.assert_called_once_with('/arb/path')
    mocked_path = mock_path.return_value
    mocked_path.parent.mkdir.assert_called_once_with(parents=True,
                                                     exist_ok=True)
    mock_output.assert_called_once_with(mocked_path, 'w', encoding=None)

    assert payload == 'one\ntwo\nthree'


@patch('pypyr.steps.filewrite.Path')
def test_filewrite_pass_with_non_string_substitutions(mock_path):
    """Convert non-string to str on text write with substitution."""
    context = Context({
        'k1': 'v1',
        'p': '/arb/path',
        'intkey': 123,
        'is_bin': False,
        'is_append': 0,
        'fileWrite': {
            'path': '{p}',
            'payload': '{intkey}',
            'binary': '{is_bin}',
            'append': '{is_append}'
        }})

    with io.StringIO() as out_text:
        with patch('pypyr.steps.filewrite.open',
                   mock_open()) as mock_output:
            mock_output.return_value.write.side_effect = out_text.write
            filewrite.run_step(context)

        payload = out_text.getvalue()

    mock_path.assert_called_once_with('/arb/path')
    mocked_path = mock_path.return_value
    mocked_path.parent.mkdir.assert_called_once_with(parents=True,
                                                     exist_ok=True)
    mock_output.assert_called_once_with(mocked_path, 'w', encoding=None)

    assert payload == '123'
# endregion substitutions



"""fileformattoml.py unit tests."""
import pytest

from pypyr.context import Context
from pypyr.errors import KeyInContextHasNoValueError, KeyNotInContextError
import pypyr.steps.fileformattoml as fileformat

# region validation


def test_fileformattoml_no_in_obj_raises():
    """None in path raises."""
    context = Context({
        'k1': 'v1'})

    with pytest.raises(KeyNotInContextError) as err_info:
        fileformat.run_step(context)

    assert str(err_info.value) == (
        "context['fileFormatToml'] doesn't exist. "
        "It must exist for pypyr.steps.fileformattoml.")


def test_fileformattoml_no_inpath_raises():
    """None in path raises."""
    context = Context({
        'fileFormatToml': 'v1'})

    with pytest.raises(KeyNotInContextError) as err_info:
        fileformat.run_step(context)

    assert str(err_info.value) == (
        "context['fileFormatToml']['in'] doesn't exist. It must exist for "
        "pypyr.steps.fileformattoml.")


def test_fileformattoml_empty_inpath_raises():
    """Empty in path raises."""
    context = Context({
        'fileFormatToml': {'in': None}})

    with pytest.raises(KeyInContextHasNoValueError) as err_info:
        fileformat.run_step(context)

    assert str(err_info.value) == ("context['fileFormatToml']['in'] must have "
                                   "a value for pypyr.steps.fileformattoml.")


# endregion validation

# region functional tests
def test_fileformattoml_pass_no_substitutions(fs):
    """Relative path to file should succeed."""
    in_path = './tests/testfiles/test.toml'
    fs.create_file(in_path, contents="""key1 = "value1"
key2 = "value2"
key3 = "value3"
""", encoding='utf-8')

    context = Context({
        'ok1': 'ov1',
        'fileFormatToml': {'in': in_path,
                           'out': './tests/testfiles/out/out.toml'}})

    fileformat.run_step(context)

    assert context, "context shouldn't be None"
    assert len(context) == 2, "context should have 2 items"
    assert context['ok1'] == 'ov1'
    assert context['fileFormatToml'] == {
        'in': in_path,
        'out': './tests/testfiles/out/out.toml'}

    with open('./tests/testfiles/out/out.toml') as outfile:
        outcontents = outfile.read()

    assert outcontents == """key1 = "value1"
key2 = "value2"
key3 = "value3"
"""


def test_fileformattoml_pass_to_out_dir(fs):
    """Relative path to file succeed with out dir rather than full path."""
    in_path = './tests/testfiles/test.toml'
    fs.create_file(in_path, contents="""key1 = "value1"
key2 = "value2"
key3 = "value3"
""", encoding='utf-8')

    context = Context({
        'ok1': 'ov1',
        'fileFormatToml': {'in': in_path,
                           'out': './tests/testfiles/out/'}})

    fileformat.run_step(context)

    assert context, "context shouldn't be None"
    assert len(context) == 2, "context should have 2 items"
    assert context['ok1'] == 'ov1'
    assert context['fileFormatToml'] == {
        'in': in_path,
        'out': './tests/testfiles/out/'}

    with open('./tests/testfiles/out/test.toml', encoding='utf-8') as outfile:
        outcontents = outfile.read()

    assert outcontents == """key1 = "value1"
key2 = "value2"
key3 = "value3"
"""


def test_fileformattoml_edit_no_substitutions(fs):
    """Relative path to file should succeed, no out means in place edit."""
    in_path = './tests/testfiles/out/edittest.toml'
    fs.create_file(in_path, contents="""key1 = "value1"
key2 = "value2"
key3 = "value3"
""", encoding='utf-8')

    context = Context({
        'ok1': 'ov1',
        'fileFormatToml': {'in': './tests/testfiles/out/edittest.toml'}})

    fileformat.run_step(context)

    assert context, "context shouldn't be None"
    assert len(context) == 2, "context should have 2 items"
    assert context['ok1'] == 'ov1'
    assert context['fileFormatToml'] == {
        'in': './tests/testfiles/out/edittest.toml'}

    with open('./tests/testfiles/out/edittest.toml',
              encoding='utf-8') as outfile:
        outcontents = outfile.read()

    assert outcontents == """key1 = "value1"
key2 = "value2"
key3 = "value3"
"""


def test_fileformattoml_pass_with_substitutions(fs):
    """Relative path to file should succeed."""
    in_path = './tests/testfiles/testsubst.toml'
    fs.create_file(in_path, contents="""key1 = "{k1}value !£$% *"
["key2_{k2}"]
abc = "{k3} def {k4}"
def = [
  "l1",
  "l2 {k5}",
  "l3",
]
k21 = "value"
""", encoding='utf-8')
    context = Context({
        'k1': 'v1',
        'k2': 'v2',
        'k3': 'v3',
        'k4': 'v4',
        'k5': 'v5',
        'fileFormatToml': {'in': in_path,
                           'out': './tests/testfiles/out/outsubst.toml'}})

    fileformat.run_step(context)

    assert context, "context shouldn't be None"
    assert len(context) == 6, "context should have 6 items"
    assert context['k1'] == 'v1'
    assert context['fileFormatToml'] == {
        'in': './tests/testfiles/testsubst.toml',
        'out': './tests/testfiles/out/outsubst.toml'}

    with open('./tests/testfiles/out/outsubst.toml',
              encoding='utf-8') as outfile:
        outcontents = outfile.read()

    expected = """key1 = "v1value !£$% *"

[key2_v2]
abc = "v3 def v4"
def = [
    "l1",
    "l2 v5",
    "l3",
]
k21 = "value"
"""

    assert outcontents == expected


def test_fileformattoml_pass_with_path_substitutions(fs):
    """Relative path to file should succeed with path substitutions."""
    in_path = './tests/testfiles/testsubst.toml'
    fs.create_file(in_path, contents="""key1 = "{k1}value !£$% *"
["key2_{k2}"]
abc = "{k3} def {k4}"
def = [
  "l1",
  "l2 {k5}",
  "l3",
]
k21 = "value"
""", encoding='utf-8')

    context = Context({
        'k1': 'v1',
        'k2': 'v2',
        'k3': 'v3',
        'k4': 'v4',
        'k5': 'v5',
        'pathIn': 'testsubst',
        'pathOut': 'outsubst',
        'fileFormatToml': {'in': './tests/testfiles/{pathIn}.toml',
                           'out': './tests/testfiles/out/{pathOut}.toml'}})

    fileformat.run_step(context)

    assert context, "context shouldn't be None"
    assert len(context) == 8, "context should have 8 items"
    assert context['k1'] == 'v1'
    assert context['fileFormatToml'] == {
        'in': './tests/testfiles/{pathIn}.toml',
        'out': './tests/testfiles/out/{pathOut}.toml'}

    with open('./tests/testfiles/out/outsubst.toml',
              encoding='utf-8') as outfile:
        outcontents = outfile.read()

    expected = """key1 = "v1value !£$% *"

[key2_v2]
abc = "v3 def v4"
def = [
    "l1",
    "l2 v5",
    "l3",
]
k21 = "value"
"""

    assert outcontents == expected


def test_fileformattoml_with_encoding(fs):
    """Toml parses binary for utf-8 only, no encoding allowed."""
    in_path = './tests/testfiles/test.toml'
    fs.create_file(in_path, contents='key1 = "value"')

    context = Context({
        'ok1': 'ov1',
        'fileFormatToml': {'in': in_path,
                           'out': 'test/out/output.toml',
                           'encoding': 'utf-16'}})

    with pytest.raises(ValueError) as err:
        fileformat.run_step(context)

    assert str(err.value) == "binary mode doesn't take an encoding argument"

# endregion functional tests



"""pypyr step that executes a string as python.

Uses python's exec() to evaluate and execute arbitrary python code.
"""
import builtins
import logging

# logger means the log level will be set correctly
logger = logging.getLogger(__name__)


def run_step(context):
    """Execute dynamic python code.

    Takes two forms of input:
        py: exec contents as dynamically interpreted python statements, with
            contents of context available as vars.
        pycode: exec contents as dynamically interpreted python statements,
            with the context object itself available as a var.

    Args:
        context (pypyr.context.Context): Mandatory.
            Context is a dictionary or dictionary-like.
            Context must contain key 'py' or 'pycode'
    """
    logger.debug("started")

    if 'pycode' in context:
        exec_pycode(context)
    else:
        context.assert_key_has_value(key='py', caller=__name__)

        # .copy() is significantly faster than globals = dict(context)
        # https://bugs.python.org/issue31179
        globals = context.copy()
        globals['__builtins__'] = builtins.__dict__

        # the save function ref allows pipeline to use save to persist vars
        # back to context,
        globals['save'] = get_save(context, globals)

        exec(context['py'], globals)

    logger.debug("done")


def exec_pycode(context):
    """Exec contents of pycode.

    This form of execute means pycode does not have the contents of context in
    the exec namespace, so referencing context needs to do:
        a = context['myvar']

    Rather than just
        a = myvar

    Args:
        context (pypyr.context.Content): context containing `pycode` key.

    Returns:
        None. Any mutations to content is on the input arg instance itself.
    """
    context.assert_key_has_value(key='pycode', caller=__name__)

    logger.debug("Executing python string: %s", context['pycode'])
    exec(context['pycode'], {'__builtins__': builtins.__dict__,
                             'context': context})


def get_save(context, namespace):
    """Return save function reference."""
    def save(*args, **kwargs):
        """Save variables in exec namespace back to context.

        Args:
            context: instance of context to which to save vars from namespace.
            namespace: set var values from this namespace.

        Returns:
            None. Mutates context.
        """
        d = {}
        for arg in args:
            try:
                d[arg] = namespace[arg]
            except KeyError as err:
                raise KeyError(f"Trying to save '{arg}', but can't find it "
                               "in the py step scope. Remember it should be "
                               "save('key'), not save(key) - mind the "
                               "quotes.") from err

        # kwargs is {} if not set, no None worries.
        d.update(**kwargs)
        context.update(d)

    return save



"""__main__.py unit tests."""
import pypyr.__main__
from unittest.mock import patch


def test_main_calls_main():
    """The python -m entry point should call pypyr.cli.main()."""
    with patch('pypyr.cli.main') as mock_cli_main:
        pypyr.__main__.main()

    mock_cli_main.assert_called_once()



"""retries.py unit tests."""
import inspect
from unittest.mock import call, patch
import pypyr.retries

# region BackoffBase


class ArbTestBackoff(pypyr.retries.BackoffBase):
    """Test backoff derivation."""

    def __call__(self, n):
        """Just return n."""
        return n


def test_backoffbase_ctor_defaults():
    """Backoff Base instantiates with defaults."""
    b = ArbTestBackoff()
    assert b.sleep == 1
    assert b.max_sleep is None
    assert b.jrc == 0
    assert b.kwargs is None
    assert b(123) == 123


def test_backoffbase_ctor_all():
    """Backoff Base instantiates with all values."""
    b = ArbTestBackoff(sleep=123, max_sleep=456, jrc=789, kwargs={'a': 'b'})
    assert b.sleep == 123
    assert b.max_sleep == 456
    assert b.jrc == 789
    assert b.kwargs == {'a': 'b'}
    assert b(123) == 123


def test_backoffbase_min_none():
    """Min when max_sleep None."""
    b = ArbTestBackoff()
    assert b.min(1) == 1
    assert b.min(4) == 4


def test_backoffbase_min():
    """Min with value for max_sleep."""
    b = ArbTestBackoff(max_sleep=3)
    assert b.min(1) == 1
    assert b.min(4) == 3


@patch('pypyr.retries.random.uniform', return_value=123)
def test_backoffbase_randomize(mock_random):
    """Randomize uses product of jrc and sleep as one side bound for random."""
    b = ArbTestBackoff(jrc=0.5)
    assert b.randomize(3) == 123

    mock_random.assert_called_once_with(1.5, 3)


@patch('pypyr.retries.random.uniform', return_value=123)
def test_backoffbase_randomize_jrc_gt_1(mock_random):
    """Randomize can use jrc > 1."""
    b = ArbTestBackoff(jrc=2)
    assert b.randomize(3) == 123

    mock_random.assert_called_once_with(6, 3)


@patch('pypyr.retries.random.uniform', return_value=123)
def test_backoffbase_randomize_default_0(mock_random):
    """Randomize uses 0 as lower bound by default."""
    b = ArbTestBackoff()
    assert b.randomize(3) == 123

    mock_random.assert_called_once_with(0, 3)

# endregion BackoffBase

# region retry strategies
# region fixed


def test_retries_fixed_with_single():
    """Fixed with single float input."""
    f = pypyr.retries.fixed(sleep=123)
    assert f(0) == 123
    assert f(1) == 123
    assert f(2) == 123


def test_retries_fixed_with_list():
    """Fixed with list input."""
    f = pypyr.retries.fixed(sleep=[2, 4, 6])
    assert f(0) == 2
    assert f(1) == 4
    assert f(2) == 6
    assert f(3) == 6
    assert f(4) == 6
    # arb repeat of n
    assert f(0) == 6


def test_retries_fixed_with_set():
    """Fixed with list input."""
    f = pypyr.retries.fixed(sleep={2, 4, 6})
    assert f(0) == 2
    assert f(1) == 4
    assert f(2) == 6
    assert f(3) == 6
    assert f(4) == 6
    # arb repeat of n
    assert f(0) == 6


def test_retries_fixed_single_with_max():
    """Fixed with single float input works with max."""
    f = pypyr.retries.fixed(sleep=123, max_sleep=456)
    assert f(0) == 123
    assert f(1) == 123
    assert f(2) == 123

    f = pypyr.retries.fixed(sleep=456, max_sleep=321)
    assert f(0) == 321
    assert f(1) == 321
    assert f(2) == 321


def test_retries_fixed_with_list_max():
    """Fixed with list input works with max."""
    f = pypyr.retries.fixed(sleep=[2, 4, 6], max_sleep=4.5)
    assert f(0) == 2
    assert f(1) == 4
    assert f(2) == 4.5
    assert f(3) == 4.5
    assert f(4) == 4.5
    # arb repeat of n
    assert f(0) == 4.5

# endregion fixed

# region jitter


@patch('pypyr.retries.random.uniform', return_value=999)
def test_retries_jitter_single(mock_random):
    """Jitter with single float input."""
    j = pypyr.retries.jitter(sleep=123)
    assert j(0) == 999
    assert j(1) == 999
    assert j(2) == 999

    assert mock_random.mock_calls == [call(0, 123),
                                      call(0, 123),
                                      call(0, 123)]


@patch('pypyr.retries.random.uniform', return_value=999)
def test_retries_jitter_single_max(mock_random):
    """Jitter with single float input honors max."""
    j = pypyr.retries.jitter(sleep=123, max_sleep=111)
    assert j(0) == 999
    assert j(1) == 999
    assert j(2) == 999

    assert mock_random.mock_calls == [call(0, 111),
                                      call(0, 111),
                                      call(0, 111)]


@patch('pypyr.retries.random.uniform', return_value=999)
def test_retries_jitter_list(mock_random):
    """Jitter with list of float input."""
    j = pypyr.retries.jitter(sleep=[1, 2, 3])
    assert j(0) == 999
    assert j(1) == 999
    assert j(2) == 999
    assert j(3) == 999
    assert j(4) == 999
    # arb repeat
    assert j(1) == 999

    assert mock_random.mock_calls == [call(0, 1),
                                      call(0, 2),
                                      call(0, 3),
                                      call(0, 3),
                                      call(0, 3),
                                      call(0, 3)]


@patch('pypyr.retries.random.uniform', return_value=999)
def test_retries_jitter_list_max(mock_random):
    """Jitter with list of float input honors max."""
    j = pypyr.retries.jitter(sleep=[1, 2, 3], max_sleep=2)
    assert j(0) == 999
    assert j(1) == 999
    assert j(2) == 999
    assert j(3) == 999
    assert j(4) == 999
    # arb repeat
    assert j(1) == 999

    assert mock_random.mock_calls == [call(0, 1),
                                      call(0, 2),
                                      call(0, 2),
                                      call(0, 2),
                                      call(0, 2),
                                      call(0, 2)]


@patch('pypyr.retries.random.uniform', return_value=999)
def test_retries_jitter_single_jrc_down(mock_random):
    """Jitter with single float input and jrc < 1."""
    j = pypyr.retries.jitter(sleep=100, jrc=0.1)
    assert j(0) == 999
    assert j(1) == 999
    assert j(2) == 999

    assert mock_random.mock_calls == [call(10, 100),
                                      call(10, 100),
                                      call(10, 100)]


@patch('pypyr.retries.random.uniform', return_value=999)
def test_retries_jitter_single_jrc_up(mock_random):
    """Jitter with single float input and jrc > 1."""
    j = pypyr.retries.jitter(sleep=100, jrc=3.5)
    assert j(0) == 999
    assert j(1) == 999
    assert j(2) == 999

    assert mock_random.mock_calls == [call(350, 100),
                                      call(350, 100),
                                      call(350, 100)]


@patch('pypyr.retries.random.uniform', return_value=999)
def test_retries_jitter_list_jrc_up_max(mock_random):
    """Jitter with list float input and jrc > 1 ignore max."""
    j = pypyr.retries.jitter(sleep=[100, 200, 300], jrc=2, max_sleep=200)
    assert j(0) == 999
    assert j(1) == 999
    assert j(2) == 999
    assert j(1) == 999

    assert mock_random.mock_calls == [call(200, 100),
                                      call(400, 200),
                                      call(400, 200),
                                      call(400, 200)]
# endregion jitter

# region linear


def test_retries_linear():
    """Linear retry works."""
    lr = pypyr.retries.linear()
    assert lr(0) == 0
    assert lr(1) == 1
    assert lr(2) == 2
    assert lr(3) == 3
    assert lr(1) == 1


def test_retries_linear_with_max():
    """Linear retry works."""
    lr = pypyr.retries.linear(sleep=2, max_sleep=8)
    assert lr(0) == 0
    assert lr(1) == 2
    assert lr(2) == 4
    assert lr(3) == 6
    assert lr(4) == 8
    assert lr(5) == 8
    assert lr(10) == 8
    assert lr(1) == 2

# endregion linear


# region linearjitter

@patch('pypyr.retries.random.uniform', return_value=999)
def test_retries_linearjitter(mock_random):
    """Linearjitter linear progression."""
    lj = pypyr.retries.linearjitter(sleep=3)
    assert lj(1) == 999
    assert lj(2) == 999
    assert lj(3) == 999
    assert lj(1) == 999

    assert mock_random.mock_calls == [call(0, 3),
                                      call(0, 6),
                                      call(0, 9),
                                      call(0, 3)]


@patch('pypyr.retries.random.uniform', return_value=999)
def test_retries_linearjitter_max(mock_random):
    """Linearjitter linear progression honors max."""
    lj = pypyr.retries.linearjitter(sleep=4, max_sleep=15)
    assert lj(1) == 999
    assert lj(2) == 999
    assert lj(3) == 999
    assert lj(4) == 999
    assert lj(2) == 999

    assert mock_random.mock_calls == [call(0, 4),
                                      call(0, 8),
                                      call(0, 12),
                                      call(0, 15),
                                      call(0, 8)]


@patch('pypyr.retries.random.uniform', return_value=999)
def test_retries_linearjitter_jrc_down(mock_random):
    """Linearjitter linear progression with jrc < 1."""
    lj = pypyr.retries.linearjitter(sleep=3, jrc=0.5)
    assert lj(1) == 999
    assert lj(2) == 999
    assert lj(3) == 999
    assert lj(1) == 999

    assert mock_random.mock_calls == [call(1.5, 3),
                                      call(3, 6),
                                      call(4.5, 9),
                                      call(1.5, 3)]


@patch('pypyr.retries.random.uniform', return_value=999)
def test_retries_linearjitter_jrc_up(mock_random):
    """Linearjitter linear progression with jrc > 1."""
    lj = pypyr.retries.linearjitter(sleep=3, jrc=2)
    assert lj(1) == 999
    assert lj(2) == 999
    assert lj(3) == 999
    assert lj(1) == 999

    assert mock_random.mock_calls == [call(6, 3),
                                      call(12, 6),
                                      call(18, 9),
                                      call(6, 3)]


@patch('pypyr.retries.random.uniform', return_value=999)
def test_retries_linearjitter_jrc_over_max(mock_random):
    """Linearjitter linear progression with jrc > 1 ignores max."""
    lj = pypyr.retries.linearjitter(sleep=3, jrc=2, max_sleep=10)
    assert lj(1) == 999
    assert lj(2) == 999
    assert lj(3) == 999
    assert lj(1) == 999

    assert mock_random.mock_calls == [call(6, 3),
                                      call(12, 6),
                                      call(18, 9),
                                      call(6, 3)]
# endregion linearjitter


# region exponential

def test_retries_exponential():
    """Exponential retry works."""
    e = pypyr.retries.exponential()
    assert e(0) == 1
    assert e(1) == 2
    assert e(2) == 4
    assert e(3) == 8
    assert e(4) == 16
    assert e(1) == 2


def test_retries_exponential_kwargs_no_base():
    """Exponential retry works with kwargs and no base set."""
    e = pypyr.retries.exponential(kwargs={'notbase': 3})
    assert e(0) == 1
    assert e(1) == 2
    assert e(2) == 4
    assert e(3) == 8
    assert e(4) == 16
    assert e(1) == 2


def test_retries_exponential_with_sleep():
    """Exponential retry works with sleep coefficient."""
    e = pypyr.retries.exponential(sleep=2)
    assert e(0) == 2
    assert e(1) == 4
    assert e(2) == 8
    assert e(3) == 16
    assert e(4) == 32
    assert e(1) == 4


def test_retries_exponential_base_3():
    """Exponential retry works with different base."""
    e = pypyr.retries.exponential(kwargs={'base': 3})
    assert e(0) == 1
    assert e(1) == 3
    assert e(2) == 9
    assert e(3) == 27
    assert e(4) == 81
    assert e(1) == 3


def test_retries_exponential_with_max():
    """Exponential retry works with max."""
    e = pypyr.retries.exponential(max_sleep=15)
    assert e(0) == 1
    assert e(1) == 2
    assert e(2) == 4
    assert e(3) == 8
    assert e(4) == 15
    assert e(5) == 15
    assert e(1) == 2

# endregion exponential

# region exponentialjitter


@patch('pypyr.retries.random.uniform', return_value=999)
def test_retries_exponentialjitter(mock_random):
    """Exponentialjitter retry works."""
    ej = pypyr.retries.exponentialjitter()
    assert ej(0) == 999
    assert ej(1) == 999
    assert ej(2) == 999
    assert ej(3) == 999
    assert ej(4) == 999
    assert ej(1) == 999
    assert ej(8) == 999

    assert mock_random.mock_calls == [call(0, 1),
                                      call(0, 2),
                                      call(0, 4),
                                      call(0, 8),
                                      call(0, 16),
                                      call(0, 2),
                                      call(0, 256)]


@patch('pypyr.retries.random.uniform', return_value=999)
def test_retries_exponentialjitter_jrc_down(mock_random):
    """Exponentialjitter retry works with jrc lower range."""
    ej = pypyr.retries.exponentialjitter(jrc=0.5)
    assert ej(0) == 999
    assert ej(1) == 999
    assert ej(2) == 999
    assert ej(3) == 999
    assert ej(4) == 999
    assert ej(1) == 999
    assert ej(8) == 999

    assert mock_random.mock_calls == [call(0.5, 1),
                                      call(1, 2),
                                      call(2, 4),
                                      call(4, 8),
                                      call(8, 16),
                                      call(1, 2),
                                      call(128, 256)]


@patch('pypyr.retries.random.uniform', return_value=999)
def test_retries_exponentialjitter_jrc_up_max(mock_random):
    """Exponentialjitter retry works with jrc upper range."""
    ej = pypyr.retries.exponentialjitter(jrc=2)
    assert ej(0) == 999
    assert ej(1) == 999
    assert ej(2) == 999
    assert ej(3) == 999
    assert ej(4) == 999
    assert ej(1) == 999
    assert ej(8) == 999

    assert mock_random.mock_calls == [call(2, 1),
                                      call(4, 2),
                                      call(8, 4),
                                      call(16, 8),
                                      call(32, 16),
                                      call(4, 2),
                                      call(512, 256)]
# endregion exponentialjitter

# endregion retry strategies


def test_backoff_export_list():
    """The backoff export list should have all BackOffBase children.

    If you've added a new back-off strategy, add it to the look-up dict:
    pypyr.retries.builtin_backoffs.
    """
    members = inspect.getmembers(
        pypyr.retries,
        lambda m: inspect.isclass(m)
        and issubclass(m, pypyr.retries.BackoffBase))

    assert pypyr.retries.builtin_backoffs == {
        n[0]: n[1] for n in members if n[0] != 'BackoffBase'}



"""pypyr step that runs another pipeline from within the current pipeline."""
from collections import namedtuple
import logging
import shlex

from pypyr.context import Context
from pypyr.errors import (ContextError,
                          ControlOfFlowInstruction,
                          KeyInContextHasNoValueError,
                          KeyNotInContextError,
                          Stop)
from pypyr.pipeline import Pipeline

PypeArgs = namedtuple('PypeArgs', ['pipeline_name',
                                   'args',
                                   'out',
                                   'use_parent_context',
                                   'pipe_arg',
                                   'skip_parse',
                                   'raise_error',
                                   'loader',
                                   'step_groups',
                                   'success_group',
                                   'failure_group',
                                   'py_dir',
                                   'parent'
                                   ])

logger = logging.getLogger(__name__)


def run_step(context):
    """Run another pipeline from this step.

    The parent pipeline is the current, executing pipeline. The invoked, or
    child pipeline is the pipeline you are calling from this step.

    Args:
        context: dictionary-like pypyr.context.Context. context is mandatory.
                 Uses the following context keys in context:
            - pype
                - name. mandatory. str. Name of pipeline to execute. This
                  {name}.yaml must exist in the working directory/pipelines
                  dir.
                - args. optional. dict. Create the context of the called
                  pipeline with these keys & values. If args specified,
                  will not pass the parent context unless you explicitly set
                  useParentContext = True. If you do set useParentContext=True,
                  will write args into the parent context.
                - out. optional. str or dict or list. If you set args or
                  useParentContext=False, the values in out will be saved from
                  child pipeline's fresh context into the parent content upon
                  completion of the child pipeline. Pass a string for a single
                  key to grab from child context, a list of string for a list
                  of keys to grab from child context, or a dict where you map
                  'parent-key-name': 'child-key-name'.
                - pipeArg. string. optional. String to pass to the
                  context_parser - the equivalent to context arg on the
                  pypyr cli. Only used if skipParse==False.
                - raiseError. bool. optional. Defaults to True. If False, log,
                  but swallow any errors that happen during the invoked
                  pipeline execution. Swallowing means that the current/parent
                  pipeline will carry on with the next step even if an error
                  occurs in the invoked pipeline.
                - skipParse. bool. optional. Defaults to True. skip the
                  context_parser on the invoked pipeline.
                - useParentContext. optional. bool. Defaults to True. Pass the
                  current (i.e parent) pipeline context to the invoked (child)
                  pipeline.
                - loader: str. optional. Absolute name of pipeline loader
                  module. If not specified will use
                  pypyr.loaders.file.
                - groups. list of str, or str. optional. Step-Groups to run in
                  pipeline. If you specify a str, will convert it to a single
                  entry list for you.
                - success. str. optional. Step-Group to run on successful
                  pipeline completion.
                - failure. str. optional. Step-Group to run on pipeline error.
                - pyDir. Path. optional. Load custom python modules from this
                  directory. You only need to set this if your custom modules
                  do not resolve from the child pipeline's parent directory
                  AND the modules are not installed to the current python
                  environment.
                - resolveFromParent. Bool. optional. Default True. Resolve
                  pipeline_name from the current (i.e calling) pipeline's
                  parent. If the child pipeline loads with the default file
                  loader and resolveFromParent is False, it will look for the
                  child pipeline in the current working directory.
                - parent. str. optional. If resolveFromParent is True, default
                  is the calling pipeline's parent.

    If none of groups, success & failure specified, will run the default pypyr
    steps, on_success & on_failure sequence.

    If groups specified, will only run groups, without a success or failure
    sequence, unless you specifically set these also.

    You're very unlikely to need to set resolveFromParent & parent, unless you
    are using a custom loader for deep customization to how pipelines load.

    Returns:
        None

    Raises:
        pypyr.errors.KeyNotInContextError: if ['pype'] or ['pype']['name']
                                           is missing.
        pypyr.errors.KeyInContextHasNoValueError: ['pype']['name'] exists but
                                                  is empty.
    """
    logger.debug("started")

    pype_args = get_arguments(context)

    try:
        pipeline, args = Pipeline.new_pipe_and_args(
            name=pype_args.pipeline_name,
            context_args=pype_args.pipe_arg,
            parse_input=not pype_args.skip_parse,
            dict_in=pype_args.args,
            loader=pype_args.loader,
            groups=pype_args.step_groups,
            success_group=pype_args.success_group,
            failure_group=pype_args.failure_group,
            py_dir=pype_args.py_dir)

        if pype_args.use_parent_context:
            logger.info("pyping %s, using parent context.",
                        pype_args.pipeline_name)

            if args:
                logger.debug("writing args into parent context...")
                context.update(args)

            pipeline.load_and_run_pipeline(context, pype_args.parent)

        else:
            logger.info("pyping %s, without parent context.",
                        pype_args.pipeline_name)

            child_context = Context(args) if args else Context()

            pipeline.load_and_run_pipeline(child_context, pype_args.parent)

            if pype_args.out:
                write_child_context_to_parent(out=pype_args.out,
                                              parent_context=context,
                                              child_context=child_context)

        logger.info("pyped %s.", pype_args.pipeline_name)
    except (ControlOfFlowInstruction, Stop):
        # Control-of-Flow/Stop are instructions to go somewhere
        # else, not errors per se.
        raise
    except Exception as ex_info:
        # yes, yes, don't catch Exception. Have to, though, in order to swallow
        # errs if !raise_error
        logger.error("Something went wrong pyping %s. %s: %s",
                     pype_args.pipeline_name, type(ex_info).__name__, ex_info)

        if pype_args.raise_error:
            logger.debug("Raising original exception to caller.")
            raise
        else:
            logger.debug("raiseError is False. Swallowing error in %s.",
                         pype_args.pipeline_name)

    logger.debug("done")


def get_arguments(context):
    """Parse arguments for pype from context and assign default values.

    Args:
        context: pypyr.context.Context. context is mandatory.

    Returns:
        PypeArgs tuple (pipeline_name, #str
                        args, #dict
                        out, #str or dict or list
                        use_parent_context, #bool
                        pipe_arg, #str
                        skip_parse, #bool
                        raise_error, #bool
                        groups, #list of str
                        success_group, #str
                        failure_group, #str
                        py_dir, # Path-like,
                        parent, #str
                        )

    Raises:
       pypyr.errors.KeyNotInContextError: if ['pype']['name'] is missing.
       pypyr.errors.KeyInContextHasNoValueError: if ['pype']['name'] exists but
                                                 is None.
    """
    context.assert_key_has_value(key='pype', caller=__name__)
    pype = context.get_formatted('pype')

    try:
        pipeline_name = pype['name']

        if pipeline_name is None:
            raise KeyInContextHasNoValueError(
                "pypyr.steps.pype ['pype']['name'] exists but is empty.")
    except KeyError as err:
        raise KeyNotInContextError(
            "pypyr.steps.pype missing 'name' in the 'pype' context item. "
            "You need to specify the pipeline name to run another "
            "pipeline.") from err

    args = pype.get('args', None)

    if args is not None and not isinstance(args, dict):
        raise ContextError(
            "pypyr.steps.pype 'args' in the 'pype' context item "
            "must be a dict.")

    pipe_arg_string = pype.get('pipeArg', None)
    pipe_arg = shlex.split(pipe_arg_string) if pipe_arg_string else None

    if pipe_arg_string and 'skipParse' not in pype:
        skip_parse = False
    else:
        skip_parse = pype.get('skipParse', True)

    if (args or pipe_arg_string) and 'useParentContext' not in pype:
        use_parent_context = False
    else:
        use_parent_context = pype.get('useParentContext', True)

    out = pype.get('out', None)
    if out and use_parent_context:
        raise ContextError(
            "pypyr.steps.pype pype.out is only relevant if useParentContext "
            "= False. If you're using the parent context, no need to have out "
            "args since their values will already be in context. If you're "
            "NOT using parent context and you've specified pype.args, just "
            "leave off the useParentContext key and it'll default to False "
            "under the hood, or set it to False yourself if you keep it in.")

    raise_error = pype.get('raiseError', True)

    groups = pype.get('groups', None)
    if isinstance(groups, str):
        groups = [groups]

    success_group = pype.get('success', None)
    failure_group = pype.get('failure', None)

    # use same loader as parent by default
    loader_info = context.current_pipeline.pipeline_definition.info
    parent_loader = loader_info.loader

    loader = pype.get('loader', parent_loader
                      if loader_info.is_loader_cascading else None)
    py_dir = pype.get('pyDir', None)

    is_resolve_from_parent = pype.get('resolveFromParent',
                                      loader_info.is_parent_cascading)

    parent_default = (loader_info.parent
                      if is_resolve_from_parent and (loader == parent_loader)
                      else None)

    parent = pype.get('parent', parent_default)

    return PypeArgs(pipeline_name,
                    args,
                    out,
                    use_parent_context,
                    pipe_arg,
                    skip_parse,
                    raise_error,
                    loader,
                    groups,
                    success_group,
                    failure_group,
                    py_dir,
                    parent)


def write_child_context_to_parent(out, parent_context, child_context):
    """Write out keys from child to parent context.

    Args:
        out. str or dict or list. Pass a string for a single
             key to grab from child context, a list of string for a list
             of keys to grab from child context, or a dict where you map
             'parent-key-name': 'child-key-name'.
        parent_context: parent Context. destination context.
        child_context: write from this context to the parent.
    """
    if isinstance(out, str):
        save_me = {out: out}
    elif isinstance(out, list):
        save_me = {k: k for k in out}
    elif isinstance(out, dict):
        save_me = out
    else:
        raise ContextError("pypyr.steps.pype pype.out should be a string, or "
                           f"a list or a dict. Instead, it's a {type(out)}")

    for parent_key, child_key in save_me.items():
        logger.debug(
            "setting parent context %s to value from child context %s",
            parent_key,
            child_key)

        parent_context[parent_key] = child_context.get_formatted(child_key)



"""add.py unit tests."""
import pytest

from pypyr.context import Context
from pypyr.dsl import PyString
from pypyr.errors import (ContextError,
                          KeyInContextHasNoValueError,
                          KeyNotInContextError)
from pypyr.steps import add


# region validation
def test_add_no_input():
    """Context add must exist."""
    with pytest.raises(KeyNotInContextError):
        add.run_step(Context())


def test_add_not_a_dict():
    """Context add must be a dict."""
    with pytest.raises(ContextError) as err:
        add.run_step(Context({
            'add': 1}))

    assert str(err.value) == (
        "context['add'] must exist, be iterable and contain 'set' for "
        "pypyr.steps.add. argument of type 'int' is not iterable")


def test_add_no_base_object():
    """Set input must exist."""
    with pytest.raises(KeyNotInContextError) as err:
        add.run_step(Context({
            'add': {}}))

    assert str(err.value) == (
        "context['add']['set'] doesn't exist. It must exist for "
        "pypyr.steps.add.")


def test_add_no_add_me():
    """Input add_me must exist."""
    with pytest.raises(KeyNotInContextError):
        add.run_step(Context({
            'add': {
                'set': 'arbset'
            }}))


def test_add_with_empty_base():
    """Input list is empty."""
    with pytest.raises(KeyInContextHasNoValueError):
        add.run_step(Context({
            'add': {
                'set': '',
                'addMe': ''}}))


def test_add_with_none_base():
    """Input set is empty."""
    with pytest.raises(KeyInContextHasNoValueError):
        add.run_step(Context({
            'add': {
                'set': None,
                'addMe': None}}))


# endregion validation

def test_add_with_none_add():
    """Add None to set."""
    context = Context({
        'add': {
            'set': 'arbset',
            'addMe': None
        }})

    add.run_step(context)

    context['add']['addMe'] = 'one'

    add.run_step(context)

    assert context['arbset'] == {None, 'one'}
    assert len(context) == 2


def test_add_pass_with_minimal_create():
    """Create set with minimal parameters success."""
    context = Context({
        'add': {
            'set': 'arbset',
            'addMe': 1
        }})

    add.run_step(context)

    context['add']['addMe'] = 2

    add.run_step(context)

    assert context['arbset'] == {1, 2}
    assert len(context) == 2


def test_add_fail_with_create_from_list_unpack_implicit():
    """When addMe not hashable fail unless is list - unpack implicit True."""
    context = Context({
        'add': {
            'set': 'arbset',
            'addMe': [1, 2]
        }})

    add.run_step(context)

    context['add']['addMe'] = [3, 4]
    add.run_step(context)

    assert context['arbset'] == {1, 2, 3, 4}
    assert len(context) == 2


def test_add_fail_with_create_from_list_unpack_false():
    """When addMe not hashable fail unless is list - unpack explicit False."""
    context = Context({
        'add': {
            'set': 'arbset',
            'addMe': [1, 2],
            'unpack': False
        }})

    with pytest.raises(TypeError):
        add.run_step(context)


def test_add_pass_with_create_from_list_extend():
    """Create new set with extend."""
    context = Context({
        'add': {
            'set': 'arbset',
            'addMe': [1, 2],
            'unpack': True
        }})

    add.run_step(context)

    context['add']['unpack'] = False
    context['add']['addMe'] = 3

    add.run_step(context)

    assert context['arbset'] == {1, 2, 3}
    assert len(context) == 2


def test_add_pass_with_create_from_list_and_add():
    """Add lists with unpack explicitly true."""
    context = Context({
        'add': {
            'set': 'arbset',
            'addMe': [1, 2],
            'unpack': True
        }})

    add.run_step(context)

    context['add']['addMe'] = [3, 4]

    add.run_step(context)

    assert context['arbset'] == {1, 2, 3, 4}
    assert len(context) == 2


def test_add_pass_existing():
    """Add to existing set."""
    context = Context({
        'arbset': {1, 2},
        'add': {
            'set': 'arbset',
            'addMe': 3
        }})

    add.run_step(context)

    context['add']['addMe'] = 4

    add.run_step(context)

    assert context['arbset'] == {1, 2, 3, 4}
    assert len(context) == 2


def test_add_pass_with_extend_existing():
    """Extend existing set."""
    context = Context({
        'arbset': {1, 2},
        'add': {
            'set': 'arbset',
            'addMe': [3, 4],
            'unpack': True
        }})

    add.run_step(context)

    context['add']['unpack'] = False
    context['add']['addMe'] = 5

    add.run_step(context)

    assert context['arbset'] == {1, 2, 3, 4, 5}
    assert len(context) == 2


def test_add_with_set_input():
    """Input is a set, not a string key."""
    context = Context({
        'arbset': {1, 2},
        'add': {
            'set': PyString('arbset'),
            'addMe': 3
        }})

    add.run_step(context)

    context['add']['addMe'] = 4

    add.run_step(context)

    assert context['arbset'] == {1, 2, 3, 4}
    assert len(context) == 2


def test_add_with_set_input_extend():
    """Input is a set, not a string key with an extend."""
    context = Context({
        'arbset': {1, 2},
        'add': {
            'set': PyString('arbset'),
            'addMe': 3
        }})

    add.run_step(context)

    context['add']['addMe'] = [4, 5]
    context['add']['unpack'] = True

    add.run_step(context)

    assert context['arbset'] == {1, 2, 3, 4, 5}
    assert len(context) == 2


def test_add_with_formatting():
    """Use formatting expressions."""
    context = Context({
        'arbset': {1, 2},
        'addthis': 3,
        'add': {
            'set': PyString('arbset'),
            'addMe': '{addthis}'
        }})

    add.run_step(context)

    context['add']['addMe'] = 4

    add.run_step(context)

    assert context['arbset'] == {1, 2, 3, 4}
    assert len(context) == 3


def test_add_set_with_formatting_tuple():
    """Use formatting expressions to add tuple."""
    context = Context({
        'arbset': {1, 2},
        'addthis': (3, 4),
        'add': {
            'set': PyString('arbset'),
            'addMe': '{addthis}',
            'unpack': True
        }})

    add.run_step(context)

    context['add']['addMe'] = 5
    context['add']['unpack'] = False

    add.run_step(context)

    assert context['arbset'] == {1, 2, 3, 4, 5}
    assert len(context) == 3


def test_add_with_formatting_extend():
    """Use formatting expressions."""
    context = Context({
        'arbset': {1, 2},
        'addthis': [3, 4],
        'extend_me': True,
        'add': {
            'set': PyString('arbset'),
            'addMe': '{addthis}',
            'unpack': '{extend_me}'
        }})

    add.run_step(context)

    context['add']['addMe'] = [5, 6]

    add.run_step(context)

    assert context['arbset'] == {1, 2, 3, 4, 5, 6}
    assert len(context) == 4


def test_add_with_single_string():
    """Add single string."""
    context = Context({
        'arbset': {1, 2},
        'add': {
            'set': PyString('arbset'),
            'addMe': 'three'
        }})

    add.run_step(context)

    context['add']['addMe'] = 'four'

    add.run_step(context)

    assert context['arbset'] == {1, 2, 'three', 'four'}
    assert len(context) == 2


def test_add_with_strings():
    """Add multiple strings."""
    context = Context({
        'arbset': {1, 2},
        'add': {
            'set': PyString('arbset'),
            'addMe': ('three', 'four'),
            'unpack': True
        }})

    add.run_step(context)

    context['add']['unpack'] = False
    context['add']['addMe'] = 'five'

    add.run_step(context)

    assert context['arbset'] == {1, 2, 'three', 'four', 'five'}
    assert len(context) == 2


def test_add_with_strings_update():
    """Add strings with update."""
    context = Context({
        'arbset': {1, 2},
        'add': {
            'set': PyString('arbset'),
            'addMe': 'xy',
            'unpack': True
        }})

    add.run_step(context)

    context['add']['unpack'] = False
    context['add']['addMe'] = 'z'

    add.run_step(context)

    assert context['arbset'] == {1, 2, 'x', 'y', 'z'}
    assert len(context) == 2



"""Arbitrary module with a test attribute.

DON'T use me anywhere except moduleloader_test.py > test_import_visitor.

Because otherwise coverage will drop <100% since Python import machinery will
cache the module lookup and test_import_visitor explicitly tests a code-path
where a module is not found in cache.

Believe it or not, this is the least painful way of achieving it.
"""


def arb_func_in_arbmod4(arg):
    """Arbitrary function just returns input."""
    return arg



"""stepcache.py unit tests."""
import pytest
from unittest.mock import patch
from pypyr.errors import PyModuleNotFoundError
import pypyr.cache.stepcache as stepcache

# ------------------------- load_the_step -----------------------------------#


def test_load_the_step_module_not_found():
    """Module not found raises."""
    with pytest.raises(PyModuleNotFoundError):
        stepcache.load_the_step('blah-blah-unlikely-name')


def test_load_the_step_attr_not_found():
    """Attribute not found raises."""
    with pytest.raises(AttributeError):
        stepcache.load_the_step(__name__)


def test_load_the_step_ok():
    """Module with attribute returns function."""
    stepcache.step_cache.clear()
    f = stepcache.load_the_step('tests.arbpack.arbmutatingstep')
    context = {'in': 'inblah'}
    f(context)

    assert context == {'in': 'inblah',
                       'inside_step': 'arb'}
# ------------------------- END load_the_step -------------------------------#

# ------------------------- StepCache ---------------------------------------#

# ------------------------- StepCache: get_step -----------------------------#


def test_get_step():
    """Step loads ok."""
    with patch('pypyr.cache.stepcache.load_the_step') as mock:
        mock.return_value = lambda x: f"{x}test"
        f = stepcache.StepCache().get_step("arbstep")

    mock.assert_called_once_with('arbstep')
    assert f("arb") == "arbtest"
# ------------------------- StepCache: get_step -----------------------------#

# ------------------------- END StepCache -----------------------------------#



"""cmd.py unit tests."""
import logging
import subprocess
from unittest.mock import call, patch

import pytest

from pypyr.config import config
from pypyr.context import Context
from pypyr.dsl import SicString
from pypyr.errors import (ContextError,
                          KeyInContextHasNoValueError,
                          KeyNotInContextError)
from pypyr.steps.dsl.cmd import CmdStep
from pypyr.subproc import Command

from tests.common.utils import patch_logger

is_windows = config.is_windows

sp_mod_name = 'pypyr.subproc'


def get_plat(posix, windows):
    """Return windows if platform is windows, else posix."""
    return windows if is_windows else posix


def test_cmdstep_name_required():
    """Cmd Step requires name."""
    with pytest.raises(AssertionError):
        CmdStep(None, None)


def test_cmdstep_context_required():
    """Cmd Step requires context."""
    with pytest.raises(AssertionError):
        CmdStep('blah', None)


def test_cmdstep_context_cmd_required():
    """Cmd Step requires cmd in context."""
    with pytest.raises(KeyNotInContextError) as err:
        CmdStep('blah', Context({'a': 'b'}))

    assert str(err.value) == ("context['cmd'] doesn't exist. It must exist "
                              "for blah.")


def test_cmdstep_context_cmd_not_none():
    """Cmd Step requires cmd in context."""
    with pytest.raises(KeyInContextHasNoValueError) as err:
        CmdStep('blah', Context({'cmd': None}))

    assert str(err.value) == "context['cmd'] must have a value for blah."


def test_cmdstep_context_cmd_not_dict():
    """Cmd Step requires cmd in context to be a dict if not str."""
    with pytest.raises(ContextError) as err:
        CmdStep('blah', Context({'cmd': 1}))

    assert str(err.value) == (
        """blah cmd config should be either a simple string:
cmd: my-executable --arg

or a dictionary:
cmd:
  run: subdir/my-executable --arg
  cwd: ./mydir

or a list of commands:
cmd:
  - my-executable --arg
  - run: another-executable --arg value
    cwd: ../mydir/subdir""")


def test_dsl_cmd_list_must_be_str_or_dict():
    """Each list input must be a string or a dict."""
    with pytest.raises(ContextError) as err:
        CmdStep('blah', Context({'cmd': ['cmd1', 123]}))

    assert str(err.value) == ("""\
123 in blah cmd config is wrong.
Each list item should be either a simple string or a dict for expanded syntax:
cmd:
  - my-executable --arg
  - run: another-executable --arg value
    cwd: ../mydir/subdir
  - run:
      - arb-executable1 --arg value1
      - arb-executable2 --arg value2
    cwd: ../mydir/arbdir""")


def test_dsl_cmd_dict_run_must_exist():
    """Dict input run must exist."""
    with pytest.raises(ContextError) as err:
        CmdStep('blah', Context({'cmd': {'runs': 'abc'}}))

    # noqa is for line too long
    assert str(err.value) == ("""\
cmd.run doesn't exist for blah.
The input should look like this in the simplified syntax:
cmd: my-executable-here --arg1

Or in the expanded syntax:
cmd:
  run: my-executable-here --arg1

If you're passing in a list of commands, each command should be a simple string,
or a dict with a `run` entry:
cmd:
  - my-executable --arg
  - run: another-executable --arg value
    cwd: ../mydir/subdir
  - run:
      - arb-executable1 --arg value1
      - arb-executable2 --arg value2
    cwd: ../mydir/arbdir""")  # noqa: E501


def test_dsl_cmd_dict_run_must_have_value():
    """Dict input run must have value."""
    with pytest.raises(ContextError) as err:
        CmdStep('blah', Context({'cmd': {'run': ''}}))

    assert str(err.value) == ("""\
cmd.run must have a value for blah.
The `run` input should look something like this:
cmd:
  run: my-executable-here --arg1
  cwd: ./mydir/subdir

Or, `run` could be a list of commands:
cmd:
  run:
    - arb-executable1 --arg value1
    - arb-executable2 --arg value2
  cwd: ../mydir/arbdir""")


def test_dsl_cmd_must_be_str_or_list():
    """Input to cmd must be a str or a list."""
    with pytest.raises(ContextError) as err:
        cmd = CmdStep('blah', Context({'cmd': {'run': 123}}))
        cmd.run_step()

    assert str(err.value) == ("""\
123 cmd should be either a simple string:
cmd: my-executable --arg

Or in the expanded syntax, set `run` to a string:

cmd:
  run: my-executable --arg
  cwd: ./mydir

Or set run to a list of commands:
cmd:
  run:
    - my-executable --arg
    - another-executable --arg2
  cwd: ../mydir/subdir""")


def test_cmdstep_cmd_is_string():
    """Str command is always not is_save."""
    obj = CmdStep('blahname', Context({'cmd': 'blah'}))

    assert not obj.is_shell
    assert obj.logger.name == 'blahname'
    assert obj.context == Context({'cmd': 'blah'})
    assert obj.commands == [Command('blah',
                                    cwd=None,
                                    is_shell=False,
                                    is_save=False)]


def test_cmdstep_cmd_is_dict_default_save_false():
    """Dict command defaults not is_save."""
    obj = CmdStep('blahname', Context({'cmd': {'run': 'blah'}}))

    assert not obj.is_shell
    assert obj.logger.name == 'blahname'
    assert obj.context == Context({'cmd': {'run': 'blah'}})
    assert obj.commands == [Command('blah',
                                    cwd=None,
                                    is_shell=False,
                                    is_save=False)]


def test_cmdstep_cmd_is_dict_default_save_true():
    """Dict command with is_save true."""
    obj = CmdStep('blahname', Context({'cmd': {'run': 'blah',
                                               'save': True}}),
                  is_shell=False)

    assert not obj.is_shell
    assert obj.logger.name == 'blahname'
    assert obj.context == Context({'cmd': {'run': 'blah', 'save': True}})
    assert obj.commands == [Command('blah',
                                    cwd=None,
                                    is_shell=False,
                                    is_save=True)]


def test_cmdstep_cmd_is_dict_cwd():
    """Cwd assigns."""
    obj = CmdStep('blahname', Context({'cmd': {'run': 'blah',
                                               'cwd': 'pathhere'}}),
                  is_shell=False)

    assert not obj.is_shell
    assert obj.logger.name == 'blahname'
    assert obj.context == Context({'cmd': {'run': 'blah', 'cwd': 'pathhere'}})
    assert obj.commands == [Command('blah',
                                    cwd='pathhere',
                                    is_shell=False,
                                    is_save=False)]


def test_dsl_cmd_shell_override():
    """Override shell arg from dict shell input."""
    obj = CmdStep('blahname', Context({'cmd': {'run': 'blah',
                                               'cwd': 'pathhere',
                                               'shell': True,
                                               }}),
                  is_shell=False)

    assert not obj.is_shell
    assert obj.logger.name == 'blahname'
    assert len(obj.commands) == 1
    cmd = obj.commands[0]
    assert cmd.is_shell
    assert not cmd.is_save


def test_cmdstep_cmd_is_dict_cwd_none():
    """Explicit None on cwd."""
    obj = CmdStep('blahname', Context({'cmd': {'run': 'blah',
                                               'cwd': None}}))

    assert not obj.is_shell
    assert obj.logger.name == 'blahname'
    assert obj.context == Context({'cmd': {'run': 'blah', 'cwd': None}})
    assert obj.commands == [Command('blah',
                                    cwd=None,
                                    is_shell=False,
                                    is_save=False)]


def test_cmdstep_runstep_cmd_is_string_shell_false():
    """Str command is always not is_save."""
    obj = CmdStep('blahname', Context({'cmd': 'blah -blah1 --blah2'}),
                  is_shell=False)

    assert not obj.is_shell
    assert obj.logger.name == 'blahname'
    assert obj.context == Context({'cmd': 'blah -blah1 --blah2'})
    assert obj.commands == [Command('blah -blah1 --blah2',
                                    cwd=None,
                                    is_shell=False,
                                    is_save=False)]

    with patch_logger(sp_mod_name, logging.DEBUG) as mock_logger_debug:
        with patch('subprocess.run') as mock_run:
            obj.run_step()

    assert mock_logger_debug.mock_calls == [
        call('stdout & stderr inheriting from parent process.'),
        call('Processing command string: blah -blah1 --blah2')]

    expected_cmd = get_plat(['blah', '-blah1', '--blah2'],
                            'blah -blah1 --blah2')

    mock_run.assert_called_once_with(expected_cmd,
                                     cwd=None,
                                     shell=False,
                                     check=True,
                                     stdout=None,
                                     stderr=None)


def test_cmdstep_runstep_cmd_is_string_shell_false_force_no_win(monkeypatch):
    """Force not windows."""
    monkeypatch.setattr('pypyr.subproc.config._is_windows', False)

    obj = CmdStep('blahname', Context({'cmd': 'blah -blah1 --blah2'}),
                  is_shell=False)

    assert not obj.is_shell
    assert obj.logger.name == 'blahname'
    assert obj.context == Context({'cmd': 'blah -blah1 --blah2'})
    assert obj.commands == [Command('blah -blah1 --blah2',
                                    cwd=None,
                                    is_shell=False,
                                    is_save=False)]

    with patch_logger(sp_mod_name, logging.DEBUG) as mock_logger_debug:
        with patch('subprocess.run') as mock_run:
            obj.run_step()

    assert mock_logger_debug.mock_calls == [
        call('stdout & stderr inheriting from parent process.'),
        call('Processing command string: blah -blah1 --blah2')]

    mock_run.assert_called_once_with(['blah', '-blah1', '--blah2'],
                                     cwd=None,
                                     shell=False,
                                     check=True,
                                     stdout=None,
                                     stderr=None)


def test_cmdstep_runstep_cmd_is_string_formatting_shell_false():
    """Str command is always not is_save and works with formatting."""
    obj = CmdStep('blahname', Context({'k1': 'blah',
                                       'cmd': '{k1} -{k1}1 --{k1}2'}),
                  is_shell=False)

    assert not obj.is_shell
    assert obj.logger.name == 'blahname'
    assert obj.context == Context({'k1': 'blah',
                                   'cmd': '{k1} -{k1}1 --{k1}2'})
    assert obj.commands == [Command('blah -blah1 --blah2',
                                    cwd=None,
                                    is_shell=False,
                                    is_save=False)]

    with patch_logger(sp_mod_name, logging.DEBUG) as mock_logger_debug:
        with patch('subprocess.run') as mock_run:
            obj.run_step()

    assert mock_logger_debug.mock_calls == [
        call('stdout & stderr inheriting from parent process.'),
        call('Processing command string: blah -blah1 --blah2')]

    expected_cmd = get_plat(['blah', '-blah1', '--blah2'],
                            'blah -blah1 --blah2')

    mock_run.assert_called_once_with(expected_cmd,
                                     cwd=None,
                                     shell=False,
                                     check=True,
                                     stdout=None,
                                     stderr=None)


def test_cmdstep_runstep_cmd_is_string_formatting_shell_false_sic():
    """Command process special tag directive."""
    obj = CmdStep('blahname',
                  Context({'k1': 'blah',
                           'cmd': SicString('{k1} -{k1}1 --{k1}2')}),
                  is_shell=False)

    assert not obj.is_shell
    assert obj.logger.name == 'blahname'
    assert obj.context == Context({'k1': 'blah',
                                   'cmd': SicString('{k1} -{k1}1 --{k1}2')})

    assert obj.commands == [Command('{k1} -{k1}1 --{k1}2',
                                    cwd=None,
                                    is_shell=False,
                                    is_save=False)]

    with patch_logger(sp_mod_name, logging.DEBUG) as mock_logger_debug:
        with patch('subprocess.run') as mock_run:
            obj.run_step()

    assert mock_logger_debug.mock_calls == [
        call('stdout & stderr inheriting from parent process.'),
        call('Processing command string: {k1} -{k1}1 --{k1}2')]

    expected_cmd = get_plat(['{k1}', '-{k1}1', '--{k1}2'],
                            '{k1} -{k1}1 --{k1}2')

    mock_run.assert_called_once_with(expected_cmd,
                                     cwd=None,
                                     shell=False,
                                     check=True,
                                     stdout=None,
                                     stderr=None)


def test_cmdstep_runstep_cmd_is_string_shell_true():
    """Str command is always not is_save."""
    obj = CmdStep('blahname',
                  Context({'cmd': 'blah -blah1 --blah2'}),
                  is_shell=True)

    assert obj.is_shell
    assert obj.logger.name == 'blahname'
    assert obj.context == Context({'cmd': 'blah -blah1 --blah2'})
    assert obj.commands == [Command('blah -blah1 --blah2',
                                    cwd=None,
                                    is_shell=True,
                                    is_save=False)]

    with patch_logger(sp_mod_name, logging.DEBUG) as mock_logger_debug:
        with patch('subprocess.run') as mock_run:
            obj.run_step()

    assert mock_logger_debug.mock_calls == [
        call('stdout & stderr inheriting from parent process.'),
        call('Processing command string: blah -blah1 --blah2')]

    # blah is in a list because shell == false
    mock_run.assert_called_once_with('blah -blah1 --blah2',
                                     cwd=None,
                                     shell=True,
                                     check=True,
                                     stdout=None,
                                     stderr=None)


def test_cmdstep_runstep_cmd_is_string_formatting_shell_true():
    """Str command is always not is_save and works with formatting."""
    obj = CmdStep('blahname',
                  Context({'k1': 'blah',
                           'cmd': '{k1} -{k1}1 --{k1}2'}),
                  is_shell=True)

    assert obj.is_shell
    assert obj.logger.name == 'blahname'
    assert obj.context == Context({'k1': 'blah',
                                   'cmd': '{k1} -{k1}1 --{k1}2'})
    assert obj.commands == [Command('blah -blah1 --blah2',
                                    cwd=None,
                                    is_shell=True,
                                    is_save=False)]

    with patch_logger(sp_mod_name, logging.DEBUG) as mock_logger_debug:
        with patch('subprocess.run') as mock_run:
            obj.run_step()

    assert mock_logger_debug.mock_calls == [
        call('stdout & stderr inheriting from parent process.'),
        call('Processing command string: blah -blah1 --blah2')]

    # blah is a string because shell == true
    mock_run.assert_called_once_with('blah -blah1 --blah2',
                                     cwd=None,
                                     shell=True,
                                     check=True,
                                     stdout=None,
                                     stderr=None)


def test_cmdstep_runstep_cmd_is_dict_save_false_shell_false():
    """Dict command with save false and shell false."""
    obj = CmdStep('blahname', Context({'cmd': {
        'run': 'blah -blah1 --blah2'}}),
        is_shell=False)

    assert not obj.is_shell
    assert obj.logger.name == 'blahname'
    assert obj.context == Context({'cmd': {'run': 'blah -blah1 --blah2'}})
    assert obj.commands == [Command('blah -blah1 --blah2',
                                    cwd=None,
                                    is_shell=False,
                                    is_save=False)]

    with patch_logger(sp_mod_name, logging.DEBUG) as mock_logger_debug:
        with patch('subprocess.run') as mock_run:
            obj.run_step()

    assert mock_logger_debug.mock_calls == [
        call('stdout & stderr inheriting from parent process.'),
        call('Processing command string: blah -blah1 --blah2')]

    # windows is always str
    expected_cmd = get_plat(['blah', '-blah1', '--blah2'],
                            'blah -blah1 --blah2')

    mock_run.assert_called_once_with(expected_cmd,
                                     cwd=None,
                                     shell=False,
                                     check=True,
                                     stdout=None,
                                     stderr=None)


def test_cmdstep_runstep_cmd_is_dict_save_false_shell_true():
    """Dict command with save false and shell true."""
    obj = CmdStep('blahname',
                  Context({'cmd': {
                      'run': 'blah -blah1 --blah2'}}),
                  is_shell=True)

    assert obj.is_shell
    assert obj.logger.name == 'blahname'
    assert obj.context == Context({'cmd': {'run': 'blah -blah1 --blah2'}})
    assert obj.commands == [Command('blah -blah1 --blah2',
                                    cwd=None,
                                    is_shell=True,
                                    is_save=False)]

    with patch_logger(sp_mod_name, logging.DEBUG) as mock_logger_debug:
        with patch('subprocess.run') as mock_run:
            obj.run_step()

    assert mock_logger_debug.mock_calls == [
        call('stdout & stderr inheriting from parent process.'),
        call('Processing command string: blah -blah1 --blah2')]

    mock_run.assert_called_once_with('blah -blah1 --blah2',
                                     cwd=None,
                                     shell=True,
                                     check=True,
                                     stdout=None,
                                     stderr=None)


def test_cmdstep_runstep_cmd_is_dict_save_false_shell_true_cwd_formatting():
    """Dict command with save false and shell true, cwd formatting."""
    obj = CmdStep('blahname', Context({
        'k1': 'v1',
        'k2': 'v2',
        'cmd': {
            'run': 'blah -blah1 --blah2', 'cwd': '/{k1}/{k2}'}}),
        is_shell=True)

    assert obj.is_shell
    assert obj.logger.name == 'blahname'
    assert obj.context == Context({'k1': 'v1',
                                   'k2': 'v2',
                                   'cmd': {
                                       'run': 'blah -blah1 --blah2',
                                       'cwd': '/{k1}/{k2}'}})

    assert obj.commands == [Command('blah -blah1 --blah2',
                                    cwd='/v1/v2',
                                    is_shell=True,
                                    is_save=False)]

    with patch('subprocess.run') as mock_run:
        with patch_logger(sp_mod_name, logging.DEBUG) as mock_logger_debug:
            obj.run_step()

    assert mock_logger_debug.mock_calls == [
        call('stdout & stderr inheriting from parent process.'),
        call('Processing command string in dir /v1/v2: blah -blah1 --blah2')]

    mock_run.assert_called_once_with('blah -blah1 --blah2',
                                     check=True,
                                     cwd='/v1/v2',
                                     shell=True,
                                     stdout=None,
                                     stderr=None)


def test_cmdstep_runstep_cmd_is_dict_save_true_shell_false():
    """Dict command with save false and shell false."""
    context = Context({'cmd': {'run': 'blah -blah1 --blah2',
                               'save': True}})

    obj = CmdStep('blahname', context)

    assert obj.is_shell is False
    assert obj.logger.name == 'blahname'
    assert obj.context == Context({'cmd': {'run': 'blah -blah1 --blah2',
                                           'save': True}})
    assert obj.commands == [Command('blah -blah1 --blah2',
                                    is_shell=False,
                                    is_save=True)]

    with patch('subprocess.run') as mock_run:
        mock_run.return_value = subprocess.CompletedProcess(None,
                                                            0,
                                                            'std',
                                                            'err')
        with patch_logger(sp_mod_name, logging.DEBUG) as mock_logger_debug:
            with patch_logger(sp_mod_name, logging.ERROR) as mock_logger_error:
                obj.run_step()

    assert mock_logger_debug.mock_calls == [
        call('stdout & stderr inheriting from parent process.'),
        call('Processing command string: blah -blah1 --blah2')]

    mock_logger_error.assert_called_once_with('stderr: err')

    # blah is in a list because shell == false on posix.
    # windows is always str
    expected_cmd = get_plat(['blah', '-blah1', '--blah2'],
                            'blah -blah1 --blah2')

    mock_run.assert_called_once_with(expected_cmd,
                                     capture_output=True,
                                     cwd=None,
                                     encoding=None,
                                     shell=False,
                                     text=True)

    assert context['cmdOut'].returncode == 0
    assert context['cmdOut'].stdout == 'std'
    assert context['cmdOut'].stderr == 'err'


def test_cmdstep_runstep_cmd_is_dict_save_true_shell_true():
    """Dict command with save false and shell true."""
    context = Context({'cmd': {'run': 'blah -blah1 --blah2',
                               'save': True}})

    obj = CmdStep('blahname', context, is_shell=True)

    assert obj.is_shell is True
    assert obj.logger.name == 'blahname'
    assert obj.context == Context({'cmd': {'run': 'blah -blah1 --blah2',
                                           'save': True}})

    assert obj.commands == [Command('blah -blah1 --blah2',
                                    is_shell=True,
                                    is_save=True)]

    with patch('subprocess.run') as mock_run:
        mock_run.return_value = subprocess.CompletedProcess(None,
                                                            0,
                                                            'std',
                                                            None)
        with patch_logger(sp_mod_name, logging.DEBUG) as mock_logger_debug:
            with patch_logger(sp_mod_name, logging.INFO) as mock_logger_info:
                obj.run_step()

    assert mock_logger_debug.mock_calls == [
        call('stdout & stderr inheriting from parent process.'),
        call('Processing command string: blah -blah1 --blah2')]
    mock_logger_info.assert_called_once_with('stdout: std')

    # blah is in a str because shell == true
    mock_run.assert_called_once_with('blah -blah1 --blah2',
                                     capture_output=True,
                                     cwd=None,
                                     encoding=None,
                                     shell=True,
                                     text=True)

    assert context['cmdOut'].returncode == 0
    assert context['cmdOut'].stdout == 'std'
    assert context['cmdOut'].stderr is None


def test_cmdstep_runstep_cmd_is_dict_save_true_shell_true_cwd_set():
    """Dict command with save false and shell true with cwd set."""
    context = Context({'cmd': {'run': 'blah -blah1 --blah2',
                               'save': True,
                               'cwd': 'pathhere'}})

    obj = CmdStep('blahname', context, is_shell=True)

    assert obj.is_shell is True
    assert obj.logger.name == 'blahname'
    assert obj.context == Context({'cmd': {'run': 'blah -blah1 --blah2',
                                           'save': True,
                                           'cwd': 'pathhere'}})

    assert obj.commands == [Command('blah -blah1 --blah2',
                                    is_shell=True,
                                    cwd='pathhere',
                                    is_save=True)]

    with patch('subprocess.run') as mock_run:
        mock_run.return_value = subprocess.CompletedProcess(None,
                                                            0,
                                                            'std',
                                                            None)
        with patch_logger(sp_mod_name, logging.DEBUG) as mock_logger_debug:
            with patch_logger(sp_mod_name, logging.INFO) as mock_logger_info:
                obj.run_step()

    assert mock_logger_debug.mock_calls == [
        call('stdout & stderr inheriting from parent process.'),
        call('Processing command string in dir pathhere: blah -blah1 --blah2')]

    mock_logger_info.assert_called_once_with('stdout: std')

    # blah is in a str because shell is true
    mock_run.assert_called_once_with('blah -blah1 --blah2',
                                     capture_output=True,
                                     cwd='pathhere',
                                     encoding=None,
                                     shell=True,
                                     text=True)

    assert context['cmdOut'].returncode == 0
    assert context['cmdOut'].stdout == 'std'
    assert context['cmdOut'].stderr is None


def test_cmdstep_runstep_cmd_is_dict_save_true_shell_false_formatting():
    """Dict command with save false and shell false with formatting."""
    context = Context({'k1': 'blah',
                       'k2': True,
                       'cmd': {'run': '{k1} -{k1}1 --{k1}2',
                               'save': '{k2}'}})

    obj = CmdStep('blahname', context)

    assert obj.is_shell is False
    assert obj.logger.name == 'blahname'
    assert obj.context == Context({'k1': 'blah',
                                   'k2': True,
                                   'cmd': {'run': '{k1} -{k1}1 --{k1}2',
                                           'save': '{k2}'}})

    assert obj.commands == [Command('blah -blah1 --blah2', is_save=True)]

    with patch('subprocess.run') as mock_run:
        mock_run.return_value = subprocess.CompletedProcess(None,
                                                            0,
                                                            'std',
                                                            'err')
        with patch_logger(sp_mod_name, logging.DEBUG) as mock_logger_debug:
            with patch_logger(sp_mod_name,
                              logging.ERROR) as mock_logger_error:
                obj.run_step()

    assert mock_logger_debug.mock_calls == [
        call('stdout & stderr inheriting from parent process.'),
        call('Processing command string: blah -blah1 --blah2')]
    mock_logger_error.assert_called_once_with('stderr: err')

    # blah is in a list because shell == false on posix.
    # windows is always str
    expected_cmd = get_plat(['blah', '-blah1', '--blah2'],
                            'blah -blah1 --blah2')

    mock_run.assert_called_once_with(expected_cmd,
                                     capture_output=True,
                                     cwd=None,
                                     encoding=None,
                                     shell=False,
                                     text=True)

    assert context['cmdOut'].returncode == 0
    assert context['cmdOut'].stdout == 'std'
    assert context['cmdOut'].stderr == 'err'



"""json.py unit tests."""
import pypyr.parser.json
import json
import pytest


def test_json_cant_parse_from_arbitrary_string():
    """Comma delimited input string should fail."""
    with pytest.raises(json.JSONDecodeError):
        pypyr.parser.json.get_parsed_context('value 1,value 2, value3')


def test_json_parser_empty_string_empty_dict():
    """Empty input creates empty dict."""
    out = pypyr.parser.json.get_parsed_context(None)
    assert not out


def test_json_parse_ok():
    """Valid json input parses."""
    out = pypyr.parser.json.get_parsed_context(['{', '"a": "b"', '}'])
    assert out == {'a': 'b'}


def test_json_parse_not_mapping_at_root():
    """Not mapping at root level raises."""
    with pytest.raises(TypeError) as err_info:
        pypyr.parser.json.get_parsed_context(['[1,', '2,', '3]'])

    assert str(err_info.value) == (
        "json input should describe an object at the top "
        "level. You should have something like \n"
        "{\n\"key1\":\"value1\",\n\"key2\":\"value2\"\n}\n"
        "at the json top-level, not an [array] or literal.")



"""assert.py unit tests."""
import importlib
import pytest
from pypyr.context import Context
from pypyr.errors import KeyNotInContextError


# loading assert dynamically because it clashes with built-in assert
assert_step = importlib.import_module('pypyr.steps.assert')


def test_assert_raises_on_empty_context():
    """Context must exist."""
    with pytest.raises(AssertionError):
        assert_step.run_step(Context())


def test_assert_raises_on_missing_assert():
    """Assert this must exist."""
    context = Context({'k1': 'v1'})
    with pytest.raises(KeyNotInContextError):
        assert_step.run_step(context)


def test_assert_raises_on_empty_assert():
    """Assert can't be empty."""
    context = Context({'assert': None})
    with pytest.raises(AssertionError) as err_info:
        assert_step.run_step(context)

    assert str(err_info.value) == "assert None evaluated to False."


def test_assert_raises_on_empty_assertthis():
    """Assert this must not be empty."""
    context = Context({'assert': {'this': None}})
    with pytest.raises(AssertionError) as err_info:
        assert_step.run_step(context)

    assert str(err_info.value) == "assert None evaluated to False."


def test_assert_passes_on_bare_assert_true():
    """Assert bare boolean True passes."""
    context = Context({'assert': True})
    assert_step.run_step(context)


def test_assert_raises_on_bare_assert_false():
    """Assert bare boolean False raises."""
    context = Context({'assert': False})
    with pytest.raises(AssertionError) as err_info:
        assert_step.run_step(context)

    assert str(err_info.value) == "assert False evaluated to False."


def test_assert_raises_on_assertthis_false():
    """Assert this boolean False raises."""
    context = Context({'assert': {'this': False}})
    with pytest.raises(AssertionError) as err_info:
        assert_step.run_step(context)

    assert str(err_info.value) == "assert False evaluated to False."


def test_assert_passes_on_assertthis_true():
    """Assert this boolean True passes."""
    context = Context({'assert': {'this': True}})
    assert_step.run_step(context)


def test_assert_passes_on_assertthis_int():
    """Assert this int 1 is True."""
    context = Context({'assert': {'this': 1}})
    assert_step.run_step(context)


def test_assert_passes_on_assertthis_arb_int():
    """Assert this non-0 int is True."""
    context = Context({'assert': {'this': 55}})
    assert_step.run_step(context)


def test_assert_passes_on_assertthis_arb_negative_int():
    """Assert this non-0 int is True."""
    context = Context({'assert': {'this': -55}})
    assert_step.run_step(context)


def test_assert_passes_on_assertthis_float():
    """Assert this non 0 float is True."""
    context = Context({'assert': {'this': 3.5}})
    assert_step.run_step(context)


def test_assert_raises_on_assertthis_false_string():
    """Assert this arbitrary string isn't True raises."""
    context = Context({'assert': {'this': 'arb string'}})
    with pytest.raises(AssertionError) as err_info:
        assert_step.run_step(context)

    assert str(err_info.value) == "assert arb string evaluated to False."


def test_assert_raises_on_assertthis_false_int():
    """Assert this int 0 is False."""
    context = Context({'assert': {'this': 0}})
    with pytest.raises(AssertionError) as err_info:
        assert_step.run_step(context)

    assert str(err_info.value) == "assert 0 evaluated to False."


def test_assert_passes_on_assertthis_true_string():
    """Assert this boolean string to True passes."""
    context = Context({'assert': {'this': 'True'}})
    assert_step.run_step(context)


def test_assert_arb_dict():
    """Arbitrary dict evaluates as truthy."""
    context = Context({'assert': {
        'arb': 'BOOM'}})

    assert_step.run_step(context)


def test_assert_arb_empty_dict():
    """Arbitrary empty dict evaluates as truthy."""
    context = Context({'assert': {}})

    with pytest.raises(AssertionError) as err_info:
        assert_step.run_step(context)

    assert str(err_info.value) == "assert {} evaluated to False."


def test_assert_raises_on_assertequals_without_this():
    """Assert raises if equals without this."""
    context = Context({'assert': {
        'equals': 'BOOM'}})
    with pytest.raises(KeyNotInContextError) as err_info:
        assert_step.run_step(context)

    assert str(err_info.value) == (
        "you have to set assert.this to use assert.equals.")


def test_assert_raises_on_assertthis_not_equals():
    """Assert this does not equal assertEquals."""
    context = Context({'assert': {
        'this': 'boom',
        'equals': 'BOOM'}})
    with pytest.raises(AssertionError) as err_info:
        assert_step.run_step(context)

    assert str(err_info.value) == (
        "assert assert['this'] is of type "
        "str and does not equal assert['equals'] of type str.")


def test_assert_passes_on_assertthis_equals():
    """Assert this equals assertEquals."""
    context = Context({'assert': {'this': 'boom',
                                  'equals': 'boom'}})
    assert_step.run_step(context)


def test_assert_passes_on_assertthis_equals_bools():
    """Assert this equals assertEquals true bools."""
    context = Context({'assert': {'this': True,
                                  'equals': True}})
    assert_step.run_step(context)


def test_assert_passes_on_assertthis_equals_bools_false():
    """Assert this equals assertEquals false bools."""
    context = Context({'assert':
                       {'this': False,
                        'equals': False}})
    assert_step.run_step(context)


def test_assert_raises_on_assertthis_not_equals_bools():
    """Assert this does not equal assertEquals bools."""
    context = Context({'assert': {'this': True,
                                  'equals': False}})
    with pytest.raises(AssertionError) as err_info:
        assert_step.run_step(context)

    assert str(err_info.value) == (
        "assert assert['this'] is of type bool and does "
        "not equal assert['equals'] of type bool.")


def test_assert_passes_on_assertthis_equals_ints():
    """Assert this equals assertEquals true ints."""
    context = Context({'assert': {'this': 33,
                                  'equals': 33}})
    assert_step.run_step(context)


def test_assert_raises_on_assertthis_not_equals_ints():
    """Assert this does not equal assertEquals ints."""
    context = Context({'assert': {'this': 0,
                                  'equals': 23}})
    with pytest.raises(AssertionError) as err_info:
        assert_step.run_step(context)

    assert str(err_info.value) == (
        "assert assert['this'] is of type int and does "
        "not equal assert['equals'] of type int.")


def test_assert_passes_on_assertthis_equals_floats():
    """Assert this equals assertEquals true ints."""
    context = Context({'assert': {'this': 123.45,
                                  'equals': 123.45}})
    assert_step.run_step(context)


def test_assert_raises_on_assertthis_not_equals_floats():
    """Assert this does not equal assertEquals ints."""
    context = Context({'assert': {'this': 123.45,
                                  'equals': 5.432}})
    with pytest.raises(AssertionError) as err_info:
        assert_step.run_step(context)

    assert str(err_info.value) == (
        "assert assert['this'] is of type float and "
        "does not equal assert['equals'] of type float.")


def test_assert_raises_on_assertthis_not_equals_string_to_int():
    """Assert this does not equal assertEquals string to int conversion."""
    context = Context({'assert': {'this': '23',
                                  'equals': 23}})
    with pytest.raises(AssertionError) as err_info:
        assert_step.run_step(context)

    assert str(err_info.value) == (
        "assert assert['this'] is of type str and does "
        "not equal assert['equals'] of type int.")


def test_assert_raises_on_assertthis_not_equals_string_to_bool():
    """Assert this string does not equal assertEquals bool."""
    context = Context({'assert': {'this': True,
                                  'equals': 'True'}})
    with pytest.raises(AssertionError) as err_info:
        assert_step.run_step(context)

    assert str(err_info.value) == (
        "assert assert['this'] is of type bool and does "
        "not equal assert['equals'] of type str.")


def test_assert_passes_on_assertthis_equals_lists():
    """Assert this equals assertEquals true list."""
    context = Context({'assert': {'this': [1, 2, 3, 4.5],
                                  'equals': [1, 2, 3, 4.5]}})
    assert_step.run_step(context)


def test_assert_raises_on_assertthis_not_equals_lists():
    """Assert this string does not equal assertEquals list."""
    context = Context({'assert': {'this': [1, 2, 8, 4.5],
                                  'equals': [1, 2, 3, 4.5]}})
    with pytest.raises(AssertionError) as err_info:
        assert_step.run_step(context)

    assert str(err_info.value) == (
        "assert assert['this'] is of type list and does "
        "not equal assert['equals'] of type list.")


def test_assert_passes_on_assertthis_equals_dicts():
    """Assert this equals assertEquals true dict."""
    context = Context({'assert': {
        'this': {'k1': 1, 'k2': [2, 3], 'k3': False},
        'equals': {'k1': 1, 'k2': [2, 3], 'k3': False}}})
    assert_step.run_step(context)


def test_assert_raises_on_assertthis_not_equals_dict_to_list():
    """Assert this string does not equal assertEquals dict."""
    context = Context({'assert': {'this': {'k1': 1, 'k2': [2, 3], 'k3': False},
                                  'equals': [1, 2, 3, 4.5]}})
    with pytest.raises(AssertionError) as err_info:
        assert_step.run_step(context)

    assert str(err_info.value) == (
        "assert assert['this'] is of type dict and does "
        "not equal assert['equals'] of type list.")


def test_assert_raises_on_assertthis_not_equals_dict_to_dict():
    """Assert this string does not equal assertEquals dict."""
    context = Context({'assert': {
        'this': {'k1': 1, 'k2': [2, 3], 'k3': False},
        'equals': {'k1': 1, 'k2': [2, 55], 'k3': False}}})
    with pytest.raises(AssertionError) as err_info:
        assert_step.run_step(context)

    assert str(err_info.value) == (
        "assert assert['this'] is of type dict and does "
        "not equal assert['equals'] of type dict.")


def test_assert_raises_on_assertthis_with_msg():
    """Assert this uses custom error message."""
    context = Context({'assert': {'this': False,
                                  'msg': "arb"}})
    with pytest.raises(AssertionError) as err_info:
        assert_step.run_step(context)

    assert str(err_info.value) == "arb"

# region substitutions


def test_assert_raises_on_assertthis_with_msg_substitutions():
    """Assert this uses custom error message."""
    context = Context({'k1': 'arb',
                       'assert': {'this': False,
                                  'msg': 'x{k1}x'}})
    with pytest.raises(AssertionError) as err_info:
        assert_step.run_step(context)

    assert str(err_info.value) == "xarbx"


def test_assert_passes_on_bare_assert_substitutions():
    """Assert bare boolean substitutes to int passes."""
    context = Context({'k1': 33, 'assert': '{k1}'})
    assert_step.run_step(context)


def test_assert_raises_on_bare_assert_substitutions():
    """Assert bare with substitution raises."""
    context = Context({'k1': 0, 'assert': '{k1}'})
    with pytest.raises(AssertionError) as err_info:
        assert_step.run_step(context)

    assert str(err_info.value) == "assert {k1} evaluated to False."


def test_assert_passes_on_assertthis_equals_ints_substitutions():
    """Assert this equals assertEquals true ints with substitutions."""
    context = Context({'k1': 33,
                       'k2': 33,
                       'assert': {'this': '{k1}',
                                  'equals': '{k2}'}})
    assert_step.run_step(context)


def test_assert_raises_on_assertthis_not_equals_ints_substitutions():
    """Assert this string does not equal assertEquals int."""
    context = Context({'k1': 33,
                       'k2': 34,
                       'assert': {'this': '{k1}',
                                  'equals': '{k2}'}})
    with pytest.raises(AssertionError) as err_info:
        assert_step.run_step(context)

    assert str(err_info.value) == (
        "assert assert['this'] is of type int and does "
        "not equal assert['equals'] of type int.")


def test_assert_passes_on_assertthis_not_equals_bools_substitutions():
    """Format expressions doesn't equivocate string True and bool True."""
    context = Context({'k1': True,
                       'k2': 'True',
                       'assert': {'this': '{k1}',
                                  'equals': '{k2}'}})

    with pytest.raises(AssertionError) as err_info:
        assert_step.run_step(context)

    assert str(err_info.value) == (
        "assert assert['this'] is of type bool and does "
        "not equal assert['equals'] of type str.")


def test_assert_passes_on_assertthis_not_equals_none_substitutions():
    """None equals None."""
    context = Context({'k1': None,
                       'k2': None,
                       'assert': {'this': '{k1}',
                                  'equals': '{k2}'}})
    assert_step.run_step(context)


def test_assert_passes_on_assertthis_true_substitutions():
    """Format expressions equivocates string True and bool True."""
    context = Context({'k1': True,
                       'k2': 'True',
                       'assert': {'this': '{k1}'}})
    assert_step.run_step(context)


def test_assert_raises_on_assertthis_not_equals_none_substitutions():
    """Assert this string does not equal assertEquals with a None."""
    context = Context({'k1': None,
                       'k2': 34,
                       'assert': {'this': '{k1}',
                                  'equals': '{k2}'}})
    with pytest.raises(AssertionError):
        assert_step.run_step(context)


def test_assert_raises_on_assertthis_bool_substitutions():
    """Assert this string substituted bool evaluates False."""
    context = Context({'k1': False,
                       'k2': 34,
                       'assert': {'this': '{k1}'}})
    with pytest.raises(AssertionError) as err_info:
        assert_step.run_step(context)

    assert str(err_info.value) == "assert {k1} evaluated to False."


def test_assert_raises_on_assertthis_substitutions_int():
    """Format expressions doesn't equivocates int 0 and bool True."""
    context = Context({'k1': 0,
                       'k2': 'True',
                       'assert': {'this': '{k1}'}})

    with pytest.raises(AssertionError) as err_info:
        assert_step.run_step(context)

    assert str(err_info.value) == "assert {k1} evaluated to False."


def test_assert_assertthis_int_1_is_true():
    """Format expressions equivocates int 1 and bool True."""
    context = Context({'k1': 1,
                       'k2': 'True',
                       'assert': {'this': '{k1}'}})

    assert_step.run_step(context)


def test_assert_raises_on_assertthis_none_substitutions():
    """Assert this string substituted None evaluates False."""
    context = Context({'k1': None,
                       'k2': 34,
                       'assert': {'this': '{k1}'}})
    with pytest.raises(AssertionError) as err_info:
        assert_step.run_step(context)

    assert str(err_info.value) == "assert {k1} evaluated to False."


def test_assert_passes_on_assertthis_equals_dicts_substitutions():
    """Assert this equals assertEquals true dict."""
    context = Context({'k1': 'v1',
                       'k2': 'v1',
                       'assert': {'this': {'k1': 1,
                                           'k2': [2, '{k1}'],
                                           'k3': False},
                                  'equals': {'k1': 1,
                                             'k2': [2, '{k2}'],
                                             'k3': False}}})
    assert_step.run_step(context)


def test_assert_passes_on_assertthis_equals_dict_substitutions():
    """Assert this equals assertEquals true dict."""
    context = Context({'k1': 'v1',
                       'k2': 'v1',
                       'dict1': {'k1': 1,
                                 'k2': [2, '{k1}'],
                                 'k3': False},
                       'dict2': {'k1': 1,
                                 'k2': [2, '{k1}'],
                                 'k3': False},
                       'assert': {'this': '{dict1}',
                                  'equals': '{dict2}'}})
    assert_step.run_step(context)


def test_assert_raises_on_assertthis_not_equals_dict_to_dict_substitutions():
    """Assert this string does not equal assertEquals dict."""
    context = Context({'k1': 'v1',
                       'k2': 'v2',
                       'assert': {'this': {'k1': 1,
                                           'k2': [2, '{k1}'],
                                           'k3': False},
                                  'equals': {'k1': 1,
                                             'k2': [2, '{k2}'],
                                             'k3': False}}})
    with pytest.raises(AssertionError) as err_info:
        assert_step.run_step(context)

    assert str(err_info.value) == (
        "assert assert['this'] is of type dict and does "
        "not equal assert['equals'] of type dict.")

# endregion substitutions



"""Administrate pypyr caches."""
import logging

from pypyr.cache.backoffcache import backoff_cache
from pypyr.cache.filecache import file_cache
from pypyr.cache.loadercache import loader_cache
from pypyr.cache.namespacecache import pystring_namespace_cache
from pypyr.cache.parsercache import contextparser_cache
from pypyr.cache.stepcache import step_cache

logger = logging.getLogger(__name__)


def clear_all() -> None:
    """Clear all pypyr caches."""
    logger.debug("clearing all cache...")

    backoff_cache.clear()
    file_cache.clear()
    loader_cache.clear()
    pystring_namespace_cache.clear()
    contextparser_cache.clear()
    step_cache.clear()

    logger.debug("all cache cleared.")



"""Global cache for namespace imports.

Attributes:
    pystring_namespace_cache: global instance of the namespace cache for
        PyString expressions. Use this attribute to access the cache from
        elsewhere.
"""
import logging
from sys import intern
from pypyr.cache.cache import Cache
from pypyr.moduleloader import ImportVisitor


logger = logging.getLogger(__name__)


class NamespaceCache(Cache):
    """Cache of namespace dictionaries.

    Parse source string for python import statements using AST Visitor.
    """

    def get_namespace(self, source):
        """Get cached namespace. Adds to cache if not exist.

        Args:
            source (str): String with python 'import x.y'/'from x import y'
                statements.

        Returns:
            Namespace dictionary of imported references.
        """
        logger.debug("starting")

        # source can be relatively long.
        # interning means cache dict compare obj id rather than full str parse
        interned_source = intern(source)
        namespace = self.get(
            interned_source,
            lambda: ImportVisitor().get_namespace(interned_source))

        logger.debug("done")
        return namespace


# global instance of the cache. use this to access the cache from elsewhere.
pystring_namespace_cache = NamespaceCache()



"""Error handling integration tests. Pipelines in ./tests/pipelines/errors."""
import tests.common.pipeline_runner as test_pipe_runner
# region runErrors

expected_file_name = '{0}_expected_output.txt'


def test_error_line_col_no():
    """Line + col numbers correct on error."""
    pipename = 'tests/pipelines/errors/line-col-no'
    test_pipe_runner.assert_pipeline_notify_output_is(pipename, ['done'])

# endregion runErrors

# region failure handler


def test_fail_no_handler():
    """No failure handler exits pypyr with exception."""
    pipename = 'tests/pipelines/errors/fail-no-handler'
    test_pipe_runner.assert_pipeline_raises(pipename,
                                            ValueError,
                                            'arb',
                                            ['A'])


def test_fail_handler():
    """Failure handler exits pypyr with exception."""
    pipename = 'tests/pipelines/errors/fail-handler'
    test_pipe_runner.assert_pipeline_raises(pipename,
                                            ValueError,
                                            'arb',
                                            ['A', 'B'])


def test_fail_handler_also_fails():
    """Failure handler also fails & exits pypyr with original exception."""
    pipename = 'tests/pipelines/errors/fail-handler-also-fails'
    test_pipe_runner.assert_pipeline_raises(pipename,
                                            TypeError,
                                            'O.G err',
                                            ['A'])


def test_fail_handler_stop():
    """Failure handler with Stop exits pypyr with no exception."""
    pipename = 'tests/pipelines/errors/fail-handler-stop'
    test_pipe_runner.assert_pipeline_notify_output_is(pipename, ['A'])


def test_fail_handler_stoppipeline():
    """Failure handler with StopPipeline exits pypyr with no exception."""
    pipename = 'tests/pipelines/errors/fail-handler-stoppipeline'
    test_pipe_runner.assert_pipeline_notify_output_is(pipename, ['A'])


def test_fail_handler_stoppipeline_pype_nested():
    """Failure handler can call to other pipelines and still stop.

    Parent pipeline has its own failure-handler, which can also stop.
    """
    pipename = 'tests/pipelines/errors/fail-handler-stoppipeline-pype-nested'
    test_pipe_runner.assert_pipeline_notify_output_is(
        pipename,
        ['A.parent', 'A', 'B.parent', 'C.parent'])


def test_fail_handler_stopstepgroup():
    """Failure handler with StopStepGroup exits pypyr with no exception."""
    pipename = 'tests/pipelines/errors/fail-handler-stopstepgroup'
    test_pipe_runner.assert_pipeline_notify_output_is(pipename, ['A'])


def test_fail_handler_stopstepgroup_multi():
    """Failure handler with StopStepGroup exits pypyr with no exception.

    Only calls 1st group in call.
    """
    pipename = 'tests/pipelines/errors/fail-handler-stopstepgroup-multi'
    test_pipe_runner.assert_pipeline_notify_output_is(pipename,
                                                      ['A', 'B', 'C'])


def test_fail_handler_stop_call_on_failure_fallback():
    """Failure handler with Stop exits pypyr with no exception.

    Only calls 1st group in call. Call with no failure handler
    falls back to default on_failure on pipeline level.
    """
    p = 'tests/pipelines/errors/fail-handler-stop-call-on-failure-fallback'
    test_pipe_runner.assert_pipeline_notify_output_is(p, ['A', 'B'])


def test_fail_handler_call_from_handler():
    """Failure handler can call to other stepgroups.

    Still raises at end.
    """
    pipename = 'tests/pipelines/errors/fail-handler-call-from-handler'
    test_pipe_runner.assert_pipeline_raises(pipename,
                                            AssertionError,
                                            'assert False evaluated to False.',
                                            ['A', 'B', 'C', 'D', 'E'])


def test_fail_handler_call_from_handler_nest_fail():
    """Failure handler can call to other stepgroups.

    Called step-group has its own failure handler.
    """
    p = 'tests/pipelines/errors/fail-handler-call-from-handler-nest-fail'
    test_pipe_runner.assert_pipeline_raises(p,
                                            AssertionError,
                                            'assert False evaluated to False.',
                                            ['A', 'B', 'C', 'D'])


def test_fail_handler_call_from_handler_nest_fail_handler():
    """Failure handler can call to other stepgroups.

    Called step-group has its own failure handler.

    Can StopStepGroup the called step-group. The outer/parent failure handler
    does NOT have a stop, so will still raise error.
    """
    pipename = (
        'tests/pipelines/errors/'
        'fail-handler-call-from-handler-nest-fail-handler')
    test_pipe_runner.assert_pipeline_raises(pipename,
                                            AssertionError,
                                            'assert False evaluated to False.',
                                            ['A', 'B', 'C', 'D', 'E', 'F'])


def test_fail_handler_call_from_handler_nest_fail_handler_stop():
    """Failure handler can call to other stepgroups and still stop.

    Called step-group has its own failure handler.

    Can StopStepGroup the called step-group. The outer/parent failure handler
    DOES have a stop too, so will not raise error.
    """
    pipename = ('tests/pipelines/errors/'
                'fail-handler-call-from-handler-nest-fail-handler-stop')
    test_pipe_runner.assert_pipeline_notify_output_is(
        pipename,
        ['A', 'B', 'C', 'D', 'E', 'F'])


def test_fail_handler_jump():
    """Failure handler can jump to other stepgroups and still raise.

    Even with the jump, still in error context, so without a Stop instruction,
    will quit reporting failure.
    """
    pipename = 'tests/pipelines/errors/fail-handler-jump'
    test_pipe_runner.assert_pipeline_raises(pipename,
                                            AssertionError,
                                            'assert False evaluated to False.',
                                            ['A', 'B', 'C'])


def test_fail_handler_jump_stop_outside():
    """Failure handler can jump to other stepgroups and still stop.

    Stop Pipeline in jumped to group stops entire pipeline without error.
    """
    pipename = 'tests/pipelines/errors/fail-handler-jump-stop-outside'
    test_pipe_runner.assert_pipeline_notify_output_is(pipename,
                                                      ['A', 'B', 'C'])
# endregion failure handler



"""pypyr step that writes payload out to a file."""
import logging
from pathlib import Path

from pypyr.config import config
from pypyr.utils.asserts import assert_key_exists, assert_key_is_truthy

logger = logging.getLogger(__name__)


def run_step(context):
    """Write payload to file.

    For list of available encodings, see:
    https://docs.python.org/3/library/codecs.html#standard-encodings

    Args:
        context: pypyr.context.Context. Mandatory.
                 The following context keys expected:
                - fileWrite
                    - path. mandatory. path-like. Write output file to
                      here. Will create directories in path for you.
                    - payload. optional. Write this value to output file.
                    - append. boolean. Default False. Set to True to append to
                      file if it exists already. If False will overwrite
                      existing file.
                    - binary. boolean. Default False. Set to True to write file
                      content as bytes in binary mode. Set both append & binary
                      True to append to binary file.
                    - encoding. string. Defaults None (platform default,
                      usually 'utf-8').

    Returns:
        None.

    Raises:
        pypyr.errors.KeyNotInContextError: fileWrite or
            fileWrite['path'] missing in context.
        pypyr.errors.KeyInContextHasNoValueError: fileWrite or
            fileWrite['path'] exists but is None/Empty.

    """
    logger.debug("started")
    context.assert_key_has_value('fileWrite', __name__)

    file_write = context.get_formatted('fileWrite')
    assert_key_is_truthy(obj=file_write,
                         key='path',
                         caller=__name__,
                         parent='fileWrite')

    assert_key_exists(obj=file_write,
                      key='payload',
                      caller=__name__,
                      parent='fileWrite')

    path = Path(file_write['path'])
    is_append = file_write.get('append', False)
    is_binary = file_write.get('binary', False)
    encoding = file_write.get('encoding', config.default_encoding)

    if is_binary:
        mode = 'ab' if is_append else 'wb'
        payload = file_write['payload']
    else:
        mode = 'a' if is_append else 'w'
        # if payload is str already, str(payload) is payload (same obj id)
        payload = str(file_write['payload'])

    logger.debug("opening destination file for writing: %s", path)

    path.parent.mkdir(parents=True, exist_ok=True)

    with open(path, mode, encoding=encoding) as file:
        file.write(payload)

    logger.info("formatted context & wrote to %s", path)
    logger.debug("done")



"""Unit tests for subproc.py."""
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path
from types import SimpleNamespace
from unittest.mock import patch, call
import sys

import pytest

from pypyr.errors import Error, ContextError
from pypyr.venv import EnvBuilderWithExtraDeps, VenvCreator

IS_AFTER_PY_3_9 = sys.version_info >= (3, 9)

# region VenvCreator


def get_simple_context():
    """Create a test SimpleNamespace to serve as EnvBuilder context."""
    context = SimpleNamespace()
    context.env_exec_cmd = '/python'
    context.bin_path = '/venv/bin'
    context.env_dir = '/venv'
    return context


@patch('pypyr.venv.VenvCreator.DEFAULT_USE_SYMLINKS', new=True)
def test_venv_creator_minimal_posix():
    """Create venv with minimal inputs on posix."""
    context = get_simple_context()
    with patch('pypyr.venv.EnvBuilderWithExtraDeps') as mock_builder:
        mock_builder.return_value.context = context
        vc = VenvCreator('/arb')
        vc.create()
        vc.install_dependencies()

    expected_path = str(Path('/arb').resolve())
    assert vc.DEFAULT_USE_SYMLINKS
    assert vc.path == expected_path
    assert vc.pip_extras is None
    assert vc.upgrade_pip
    assert vc._is_done
    assert vc._future is None
    mock_builder.assert_called_once_with(system_site_packages=False,
                                         clear=False,
                                         symlinks=True,
                                         upgrade=False,
                                         with_pip=True,
                                         prompt=None,
                                         upgrade_deps=False,
                                         is_quiet=False)
    mocked_builder = mock_builder.return_value
    mocked_builder.create.assert_called_once_with(expected_path)
    mocked_builder.upgrade_dependencies.assert_called_once_with(context)
    mocked_builder.pip_install_extras.assert_not_called()


@patch('pypyr.venv.VenvCreator.DEFAULT_USE_SYMLINKS', new=False)
def test_venv_creator_minimal_nt():
    """Create venv with minimal inputs on nt."""
    context = get_simple_context()
    with patch('pypyr.venv.EnvBuilderWithExtraDeps') as mock_builder:
        mock_builder.return_value.context = context
        vc = VenvCreator('/arb')
        vc.create()
        vc.install_dependencies()

    expected_path = str(Path('/arb').resolve())
    assert not vc.DEFAULT_USE_SYMLINKS
    assert vc.path == expected_path
    assert vc.pip_extras is None
    assert vc.upgrade_pip
    assert vc._is_done
    assert vc._future is None
    mock_builder.assert_called_once_with(system_site_packages=False,
                                         clear=False,
                                         symlinks=False,
                                         upgrade=False,
                                         with_pip=True,
                                         prompt=None,
                                         upgrade_deps=False,
                                         is_quiet=False)

    mocked_builder = mock_builder.return_value
    mocked_builder.create.assert_called_once_with(expected_path)
    mocked_builder.upgrade_dependencies.assert_called_once_with(context)
    mocked_builder.pip_install_extras.assert_not_called()


def test_venv_creator_from_mapping_minimal():
    """Create venv with minimal inputs from mapping."""
    d = {'path': '~/arb'}
    resolved_path = str(Path('~/arb').expanduser().resolve())

    context = get_simple_context()
    with patch('pypyr.venv.EnvBuilderWithExtraDeps') as mock_builder:
        mock_builder.return_value.context = context
        vcs = list(VenvCreator.from_mapping(d))
        assert len(vcs) == 1
        vc = vcs[0]
        vc.create()
        vc.install_dependencies()

    assert vc.path == resolved_path
    assert vc.pip_extras is None
    assert vc.upgrade_pip
    assert vc._is_done
    assert vc._future is None
    mock_builder.assert_called_once_with(
        system_site_packages=False,
        clear=False,
        symlinks=VenvCreator.DEFAULT_USE_SYMLINKS,
        upgrade=False,
        with_pip=True,
        prompt=None,
        upgrade_deps=False,
        is_quiet=False)

    mocked_builder = mock_builder.return_value
    mocked_builder.create.assert_called_once_with(resolved_path)
    mocked_builder.upgrade_dependencies.assert_called_once_with(context)
    mocked_builder.pip_install_extras.assert_not_called()


def test_venv_creator_from_mapping_minimal_list():
    """Create venv with minimal inputs from mapping with list of paths."""
    d = {'path': ['/arb1', '/arb2']}
    context = get_simple_context()
    with patch('pypyr.venv.EnvBuilderWithExtraDeps') as mock_builder:
        mock_builder.return_value.context = context
        vcs = list(VenvCreator.from_mapping(d))
        assert len(vcs) == 2
        vc0 = vcs[0]
        vc0.create()
        vc0.install_dependencies()

        vc1 = vcs[1]
        vc1.create()
        vc1.install_dependencies()

    expected_path1 = str(Path('/arb1').resolve())
    expected_path2 = str(Path('/arb2').resolve())
    assert vc0.path == expected_path1
    assert vc0.pip_extras is None
    assert vc0.upgrade_pip
    assert vc0._is_done
    assert vc0._future is None

    assert vc1.path == expected_path2
    assert vc1.pip_extras is None
    assert vc0.upgrade_pip
    assert vc1._is_done
    assert vc1._future is None

    assert mock_builder.call_count == 2
    assert mock_builder.mock_calls[0] == call(
        system_site_packages=False,
        clear=False, symlinks=VenvCreator.DEFAULT_USE_SYMLINKS,
        upgrade=False, with_pip=True,
        prompt=None, upgrade_deps=False,
        is_quiet=False)

    assert mock_builder.mock_calls[1] == call(
        system_site_packages=False,
        clear=False, symlinks=VenvCreator.DEFAULT_USE_SYMLINKS,
        upgrade=False, with_pip=True,
        prompt=None, upgrade_deps=False,
        is_quiet=False)

    mocked_builder = mock_builder.return_value
    assert mocked_builder.create.call_count == 2
    assert mocked_builder.create.mock_calls[0] == call(expected_path1)
    assert mocked_builder.create.mock_calls[1] == call(expected_path2)

    assert mocked_builder.upgrade_dependencies.call_count == 2
    assert mocked_builder.upgrade_dependencies.mock_calls[0] == call(context)
    assert mocked_builder.upgrade_dependencies.mock_calls[1] == call(context)

    mocked_builder.pip_install_extras.assert_not_called()


def test_venv_creator_from_mapping_maximal():
    """Create venv with maximal inputs from mapping."""
    d = {
        'path': '/arb',
        'system_site_packages': True,
        'clear': False,
        'symlinks': True,
        'upgrade': True,
        'with_pip': True,
        'prompt': 'arbprompt',
        'upgrade_pip': False,
        'pip': 'package1 package2',
        'quiet': True
    }

    context = get_simple_context()

    with patch('pypyr.venv.EnvBuilderWithExtraDeps') as mock_builder:
        mock_builder.return_value.context = context
        vcs = list(VenvCreator.from_mapping(d))
        assert len(vcs) == 1
        vc = vcs[0]
        vc.create()
        vc.install_dependencies()

    expected_path = str(Path('/arb').resolve())
    assert vc.path == expected_path
    assert vc.pip_extras == 'package1 package2'
    assert not vc.upgrade_pip
    assert vc._is_done
    assert vc._future is None
    mock_builder.assert_called_once_with(system_site_packages=True,
                                         clear=False,
                                         symlinks=True,
                                         upgrade=True,
                                         with_pip=True,
                                         prompt='arbprompt',
                                         upgrade_deps=False,
                                         is_quiet=True)

    mocked_builder = mock_builder.return_value
    mocked_builder.create.assert_called_once_with(expected_path)
    mocked_builder.upgrade_dependencies.assert_not_called()
    mocked_builder.pip_install_extras.assert_called_once_with(
        'package1 package2')


def test_venv_creator_from_mapping_maximal_no_pip():
    """Create venv with maximal inputs from mapping when with_pip False."""
    d = {
        'path': '/arb',
        'system_site_packages': True,
        'clear': False,
        'symlinks': True,
        'upgrade': True,
        'with_pip': False,
        'prompt': 'arbprompt',
        'upgrade_pip': False,
        'quiet': True
    }

    context = get_simple_context()

    with patch('pypyr.venv.EnvBuilderWithExtraDeps') as mock_builder:
        mock_builder.return_value.context = context
        vcs = list(VenvCreator.from_mapping(d))
        assert len(vcs) == 1
        vc = vcs[0]
        vc.create()
        vc.install_dependencies()

    expected_path = str(Path('/arb').resolve())
    assert vc.path == expected_path
    assert vc.pip_extras is None
    assert not vc.upgrade_pip
    assert vc._is_done
    assert vc._future is None
    mock_builder.assert_called_once_with(system_site_packages=True,
                                         clear=False,
                                         symlinks=True,
                                         upgrade=True,
                                         with_pip=False,
                                         prompt='arbprompt',
                                         upgrade_deps=False,
                                         is_quiet=True)

    mocked_builder = mock_builder.return_value
    mocked_builder.create.assert_called_once_with(expected_path)
    mocked_builder.upgrade_dependencies.assert_not_called()
    mocked_builder.pip_install_extras.assert_not_called()


def test_venv_creator_from_mapping_no_path_error():
    """Raise error when instantiate from mapping with no path."""
    with pytest.raises(ContextError) as err:
        next(VenvCreator.from_mapping({'a': 'b'}))

    assert str(err.value) == (
        "path is mandatory on mapping input for venv create.")


def test_venv_creator_from_mapping_path_not_string_not_list():
    """Raise error when instantiate from mapping with path wrong type."""
    with pytest.raises(ContextError) as err:
        next(VenvCreator.from_mapping({'path': 123}))

    assert str(err.value) == (
        "venv.path input should be a string or list. "
        + "Current input is: 123")


def test_venv_creator_error_upgrade_and_clear():
    """Raise error when upgrade and clear true."""
    with pytest.raises(ContextError) as err:
        VenvCreator('/arb', upgrade=True, clear=True)

    assert str(err.value) == VenvCreator.EXCEPTION_UPGRADE_CLEAR.format(
        path='/arb')


def test_venv_creator_no_pip_sets_upgrade_pip_false():
    """Set upgrade_pip False when with_pip is False."""
    vc = VenvCreator('/arb', with_pip=False)
    assert not vc.with_pip
    assert not vc.upgrade_pip

    vc = VenvCreator('/arb', with_pip=False, upgrade_pip=True)
    assert not vc.with_pip
    assert not vc.upgrade_pip


def test_venv_creator_error_no_pip_with_extras():
    """Raise error when with_pip false and pip_extras or upgrade_pip set."""
    with pytest.raises(ContextError) as err:
        VenvCreator('/arb', with_pip=False, pip_extras="package1 package2")

    assert str(err.value) == VenvCreator.EXCEPTION_NO_PIP.format(path='/arb')


def test_venv_creator_error_install_deps_before_create():
    """Raise error when install dependencies before create."""
    vc = VenvCreator('/arb')

    with pytest.raises(Error) as err:
        vc.install_dependencies()

    expected_path = str(Path('/arb').resolve())
    assert str(err.value) == (
        "you can't call install_dependencies before the environment "
        + f"is installed successfully for {expected_path}.")


@patch('pypyr.venv.EnvBuilderWithExtraDeps')
def test_venv_creator_error_install_deps_with_pip_false(mock_builder):
    """Raise error when install dependencies when with_pip is false."""
    vc = VenvCreator('/arb', with_pip=False)
    context = get_simple_context()

    with pytest.raises(Error) as err:
        mock_builder.return_value.with_pip = False
        mock_builder.return_value.context = context
        vc.create()
        vc.install_dependencies()

    expected_path = str(Path('/arb').resolve())
    assert str(err.value) == (
        f"you can't install extra dependencies into {expected_path} because "
        + "with_pip is False.")

    mock_builder.assert_called_once_with(
        system_site_packages=False,
        clear=False,
        symlinks=VenvCreator.DEFAULT_USE_SYMLINKS,
        upgrade=False,
        with_pip=False,
        prompt=None,
        upgrade_deps=False,
        is_quiet=False)

    mocked_builder = mock_builder.return_value
    mocked_builder.create.assert_called_once_with(expected_path)
    mocked_builder.upgrade_dependencies.assert_not_called()


def test_venv_creator_error_check_result_no_executor():
    """Raise error when check_result without create_in_executor.."""
    vc = VenvCreator('/arb')

    with pytest.raises(Error) as err:
        vc.check_result()

    expected_path = str(Path('/arb').resolve())
    assert str(err.value) == (
        "You can only call check_result after create_in_executor. "
        + f"Error creating venv {expected_path}")


@patch('pypyr.venv.EnvBuilderWithExtraDeps')
def test_venv_creator_in_executor(mock_builder):
    """Create venv in executor."""
    vc = VenvCreator('/arb')
    with ThreadPoolExecutor() as executor:
        vc.create_in_executor(executor)

    assert not vc.check_result()

    mock_builder.assert_called_once_with(
        system_site_packages=False,
        clear=False,
        symlinks=VenvCreator.DEFAULT_USE_SYMLINKS,
        upgrade=False,
        with_pip=True,
        prompt=None,
        upgrade_deps=False,
        is_quiet=False)

    expected_path = str(Path('/arb').resolve())

    mocked_builder = mock_builder.return_value
    mocked_builder.create.assert_called_once_with(expected_path)
    mocked_builder.upgrade_dependencies.assert_not_called()


@patch('pypyr.venv.EnvBuilderWithExtraDeps')
def test_venv_creator_in_executor_raises(mock_builder):
    """Create venv in executor where create raises error."""
    vc = VenvCreator('/arb')
    mocked_builder = mock_builder.return_value
    mocked_builder.create.side_effect = ValueError('arb')

    with ThreadPoolExecutor() as executor:
        vc.create_in_executor(executor)

    assert repr(vc.check_result()) == repr(ValueError('arb'))

    mock_builder.assert_called_once_with(
        system_site_packages=False,
        clear=False,
        symlinks=VenvCreator.DEFAULT_USE_SYMLINKS,
        upgrade=False,
        with_pip=True,
        prompt=None,
        upgrade_deps=False,
        is_quiet=False)

    expected_path = str(Path('/arb').resolve())
    mocked_builder.create.assert_called_once_with(expected_path)
    mocked_builder.upgrade_dependencies.assert_not_called()


@patch('pypyr.venv.EnvBuilderWithExtraDeps')
def test_venv_creator_in_executor_raises_base_exception(mock_builder):
    """Create venv in executor where create raises KeyBoardInterrupt."""
    vc = VenvCreator('/arb')
    mocked_builder = mock_builder.return_value
    mocked_builder.create.side_effect = KeyboardInterrupt()

    with ThreadPoolExecutor() as executor:
        vc.create_in_executor(executor)

    with pytest.raises(KeyboardInterrupt):
        vc.check_result()

    mock_builder.assert_called_once_with(
        system_site_packages=False,
        clear=False,
        symlinks=VenvCreator.DEFAULT_USE_SYMLINKS,
        upgrade=False,
        with_pip=True,
        prompt=None,
        upgrade_deps=False,
        is_quiet=False)

    expected_path = str(Path('/arb').resolve())
    mocked_builder.create.assert_called_once_with(expected_path)
    mocked_builder.upgrade_dependencies.assert_not_called()
# endregion VenvCreator

# region EnvBuilderWithContext


def test_env_builder_minimal():
    """Instantiate env builder with minimal args."""
    eb = EnvBuilderWithExtraDeps()
    assert not eb.is_quiet

    assert not eb.system_site_packages
    assert not eb.clear
    assert not eb.symlinks
    assert not eb.upgrade
    assert not eb.with_pip
    assert eb.prompt is None
    assert not eb.upgrade_deps


def test_env_builder_maximal():
    """Instantiate env builder with maximal args."""
    eb = EnvBuilderWithExtraDeps(system_site_packages=True,
                                 clear=True,
                                 symlinks=True,
                                 upgrade=False,
                                 with_pip=True,
                                 prompt='arbprompt',
                                 upgrade_deps=True,
                                 is_quiet=True)

    assert eb.is_quiet

    assert eb.system_site_packages
    assert eb.clear
    assert eb.symlinks
    assert not eb.upgrade
    assert eb.with_pip
    assert eb.prompt == 'arbprompt'
    assert eb.upgrade_deps


@patch('pypyr.venv.IS_BEFORE_PY_3_9', new=True)
def test_env_builder_maximal_before_py_3_9():
    """Instantiate env builder with maximal args before py 3.9."""
    with patch('pypyr.venv.EnvBuilder.__init__') as mock_builder:
        eb = EnvBuilderWithExtraDeps(system_site_packages=True,
                                     clear=True,
                                     symlinks=True,
                                     upgrade=False,
                                     with_pip=True,
                                     prompt='arbprompt',
                                     upgrade_deps=True,
                                     is_quiet=True)

    assert eb.is_quiet

    mock_builder.assert_called_once_with(system_site_packages=True,
                                         clear=True,
                                         symlinks=True,
                                         upgrade=False,
                                         with_pip=True,
                                         prompt='arbprompt')

    assert eb.upgrade_deps


@patch('pypyr.venv.IS_BEFORE_PY_3_9', new=False)
def test_env_builder_maximal_after_py_3_9():
    """Instantiate env builder with maximal args after py 3.9."""
    with patch('pypyr.venv.EnvBuilder.__init__') as mock_builder:
        eb = EnvBuilderWithExtraDeps(system_site_packages=True,
                                     clear=True,
                                     symlinks=True,
                                     upgrade=False,
                                     with_pip=True,
                                     prompt='arbprompt',
                                     upgrade_deps=True,
                                     is_quiet=True)

    assert eb.is_quiet

    mock_builder.assert_called_once_with(system_site_packages=True,
                                         clear=True,
                                         symlinks=True,
                                         upgrade=False,
                                         with_pip=True,
                                         prompt='arbprompt',
                                         upgrade_deps=True)


def test_env_builder_maximal_upgrade():
    """Instantiate env builder with maximal args and upgrade True."""
    eb = EnvBuilderWithExtraDeps(system_site_packages=True,
                                 clear=False,
                                 symlinks=True,
                                 upgrade=True,
                                 with_pip=True,
                                 prompt='arbprompt',
                                 upgrade_deps=True,
                                 is_quiet=True)

    assert eb.is_quiet

    assert eb.system_site_packages
    assert not eb.clear
    assert eb.symlinks
    assert eb.upgrade
    assert eb.with_pip
    assert eb.prompt == 'arbprompt'
    assert eb.upgrade_deps


@patch('pypyr.venv.subprocess.check_call')
def test_env_builder_upgrade_deps_no_quiet(mock_subproc_run):
    """Install dependencies without quiet mode."""
    eb = EnvBuilderWithExtraDeps()
    context = get_simple_context()

    eb.post_setup(context)
    eb.upgrade_dependencies(context)

    mock_subproc_run.assert_called_once_with(
        ['/python', '-m', 'pip', 'install', '--upgrade', 'pip', 'setuptools'])


@patch('pypyr.venv.subprocess.check_call')
def test_env_builder_upgrade_deps_quiet(mock_subproc_run):
    """Install dependencies with quiet mode."""
    eb = EnvBuilderWithExtraDeps(is_quiet=True)
    context = get_simple_context()

    eb.post_setup(context)
    eb.upgrade_dependencies(context)

    mock_subproc_run.assert_called_once_with(
        ['/python', '-m', 'pip', 'install', '--upgrade', '-q',
         'pip', 'setuptools'])


@patch('pypyr.venv.subprocess.run')
def test_env_builder_install_pip_list_input(mock_subproc_run):
    """Install dependencies with list input rather than string."""
    eb = EnvBuilderWithExtraDeps()
    context = get_simple_context()

    eb.post_setup(context)
    eb.pip_install_extras(
        ['package1', 'package2==1.2.3', 'package3>=4.5.6,<7.8.9'])

    mock_subproc_run.assert_called_once_with(
        ['/python', '-m', 'pip', 'install',
         'package1', 'package2==1.2.3', 'package3>=4.5.6,<7.8.9'],
        check=True)


@patch('pypyr.venv.subprocess.run')
def test_env_builder_install_pip_extras(mock_subproc_run):
    """Install dependencies without quiet mode."""
    eb = EnvBuilderWithExtraDeps()
    context = get_simple_context()

    eb.post_setup(context)
    eb.pip_install_extras('package1 package2==1.2.3 package3>=4.5.6,<7.8.9')

    mock_subproc_run.assert_called_once_with(
        ['/python', '-m', 'pip', 'install',
         'package1', 'package2==1.2.3', 'package3>=4.5.6,<7.8.9'],
        check=True)


@patch('pypyr.venv.subprocess.run')
def test_env_builder_install_pip_extras_quiet(mock_subproc_run):
    """Install dependencies with quiet mode."""
    eb = EnvBuilderWithExtraDeps(is_quiet=True)
    context = get_simple_context()

    eb.post_setup(context)
    eb.pip_install_extras('package1 package2==1.2.3 package3>=4.5.6,<7.8.9')

    mock_subproc_run.assert_called_once_with(
        ['/python', '-m', 'pip', 'install', '-q',
         'package1', 'package2==1.2.3', 'package3>=4.5.6,<7.8.9'],
        check=True)

# endregion EnvBuilderWithContext
